import cv2
import os
from PIL import Image
import matplotlib.pyplot as plt
import torch
import mmcv
import numpy as np
import time
import yaml
from mmcv import imread
import mmengine
from mmengine.registry import init_default_scope

from mmpose.apis import inference_topdown
from mmpose.apis import init_model as init_pose_estimator
from mmpose.evaluation.functional import nms
from mmpose.registry import VISUALIZERS
from mmpose.structures import merge_data_samples
from mmdet.apis import inference_detector, init_detector
from .mmpose_estimator import MMposeEstimator

class MMposeEstimatorBuilder:
    def __init__(self, file):
        assert file.endswith(('.yaml', '.yml'))
        with open(file, 'r') as f:
            setting = yaml.safe_load(f)
        self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        key = list(setting.keys())[0]
        self.detector_config = setting[key]['detector']['config_path']
        self.detector_checkpoint = setting[key]['detector']['checkpoint_path']
        self.pose_config = setting[key]['pose']['config_path']
        self.pose_checkpoint = setting[key]['pose']['checkpoint_path']
        self.draw_setting_path = setting[key]['draw_setting_path']
        self.conf_thresh = setting[key]['conf_thresh']
        self.iou_thresh = setting[key]['iou_thresh']

        self.detector = init_detector(
            self.detector_config,
            self.detector_checkpoint,
            device=self.device
        )
        self.pose_estimator = init_pose_estimator(
            self.pose_config,
            self.pose_checkpoint,
            device=self.device,
            cfg_options={'model': {'test_cfg': {'output_heatmaps': False}}}
        )
        
    def __call__(self):
        self.mmpose_estimator = MMposeEstimator(self.detector, self.pose_estimator, self.draw_setting_path, self.conf_thresh, self.iou_thresh, self.device)
        return self.mmpose_estimator

    def draw_pose_results_on_images(self, images, bboxes, keypoints, show_id=False, conf_thres=0.25):
        return self.mmpose_estimator.draw_pose_results_on_images(images, bboxes, keypoints, show_id, conf_thres)


import os
import sys
import time
import logging
import datetime
from pathlib import Path

class Logger():
    """
    A custom logger class that handles log creation, daily log rotation,
    and automatic deletion of expired log files based on a defined expiration.

    Attributes:
        level (str): The logging level for the logger (e.g., 'DEBUG', 'INFO').
        console (str): The logging level for console output.
        file (str): The logging level for file output.
        root (str): The root directory where log files are stored.
        file_prefix (str): Prefix for log files.
        expired (int): Number of days after which log files are considered expired and deleted.
        namespace (str): Namespace for the logger, useful for differentiating logs in larger applications.
        location (str): Specifies whether to log the filename or pathname in the logs.
        style (str): Style of the log format, can be '[', '-', or ':'.

    Example:
        logger = Logger(level='DEBUG', console='INFO', file='ERROR', root='logs', file_prefix='app')
        logger.info("This is an info message.")
    """

    VALID_LOG_LEVELS = ['debug', 'info', 'warning', 'error', 'critical', 'exception']

    def __init__(
            self,
            level='DEBUG',
            console='',
            file='',
            root='logs',
            file_prefix='',
            expired=-1,
            namespace='',
            location='',  # 'filename' or 'pathname' or ''
            style='-'  # '[' or '-' or ':'
        ):
        self.level = level
        self.console = console.upper() if console else ''
        self.file = file.upper() if file else ''
        if file:
            self.root = root
            self.file_prefix = file_prefix + '_' if file_prefix != '' else ''
            self.expired = expired
            self.delete_expired_log()

        self.namespace = namespace

        if location and sys.version_info < (3, 8):
            # stacklevel is only supported in Python 3.8 and above
            print("Warning: 'location' parameter is only supported in Python 3.8 and above. Setting location to False.")
            self.location = ''
        elif location not in ['filename', 'pathname', '']:
            self.location = ''
        else:
            self.location = location
        
        self.style = style
        if self.style not in ['[', '-', ':']:
            print(f"Warning: Invalid style '{self.style}' provided. Defaulting to '['.")
            self.style = '['

        self.day = time.strftime('%Y-%m-%d')
        self.setup()

    def __getattr__(self, name):
        """
        Allows dynamic access to standard logging methods like debug, info, etc.
        
        Raises:
            AttributeError: If the attribute is not a recognized log level.
        """
        if name in self.VALID_LOG_LEVELS:
            def log_method(msg, *args, **kwargs):
                self.check()
                if self.location and 'stacklevel' not in kwargs:
                    # Since now the logger is a class attribute, 
                    # we need to set stacklevel to 2
                    # in order to point to the caller of the log method.
                    kwargs['stacklevel'] = 2
                getattr(self.logger, name)(msg, *args, **kwargs)
            return log_method
        raise AttributeError(f"'{name}' is not a valid logging level. Valid levels are: {', '.join(self.VALID_LOG_LEVELS)}")
 
    def setup(self):
        """
        Create or recreate the logging handlers based on the current configuration.
        This is called internally to handle log rotation.
        """
        self.shutdown()

        # create a logger obj and set level
        # use __name__ to avoid use root logger
        if self.namespace:
            if self.namespace == 'global':
                self.logger = logging.getLogger()
            else:
                self.logger = logging.getLogger(self.namespace)
        else:
            self.logger = logging.getLogger(__name__)
        self.logger.setLevel(getattr(logging, self.level))
        
        # create a formatter obj and add it into the handlers
        self.formatter = logging.Formatter(self.get_format())
        self.formatter.datefmt = '%Y-%m-%d %H:%M:%S'
        
        # create a console handler
        if self.console:
            console_handler = logging.StreamHandler()
            console_handler.setLevel(getattr(logging, self.console))
            console_handler.setFormatter(self.formatter)
            self.logger.addHandler(console_handler)
            
        # create a file handler
        if self.file:
            if not os.path.isdir(self.root):
                os.mkdir(self.root)
            file_handler = logging.FileHandler(f"{self.root}/{self.file_prefix}{self.day}.log")
            file_handler.setLevel(getattr(logging, self.file))
            file_handler.setFormatter(self.formatter)
            self.logger.addHandler(file_handler)

    def get_format(self):
        """
        Returns the log format string based on the logger's configuration.
        """
        _format = []
        _format.append('%(asctime)s')
        _format.append('%(levelname)s')
        if self.location:
            # location can be 'filename', 'pathname', or ''
            # %(filename)s is the name of the file without the path
            # %(pathname)s is the full path of the file
            # %(lineno)d is the line number in the file where the log was called
            _format.append(f'%({self.location})s:%(lineno)d')
        _format.append('%(message)s')

        if self.style == '[':
            _format = [f'[{part}]' for part in _format[:-1]] + [_format[-1]]
            return ' '.join(_format)
        elif self.style == '-':
            return ' - '.join(_format)
        elif self.style == ':':
            return ':'.join(_format)
    
    def delete_expired_log(self):
        """
        Delete log files that are older than the 'expired' days setting.
        """
        if not self.file:
            return
        
        if self.expired < 0:
            return
        
        today = datetime.date.today()
        log_dir = Path(self.root)
        if not log_dir.exists():
            return

        for file_path in log_dir.iterdir():
            if file_path.suffix == '.log':
                try:
                    file_date_str = file_path.stem.split('_')[-1]
                    file_date = datetime.datetime.strptime(file_date_str, "%Y-%m-%d").date()
                    if (today - file_date).days > self.expired:
                        file_path.unlink()
                        print(f'delete expired log: {file_path}', flush=True)
                except ValueError:
                    print(f"Skipped invalid log file: {file_path}", flush=True)
                except Exception as e:
                    print(f"Error deleting file {file_path}: {e}", flush=True)

    def check(self):
        """
        Check if the log file needs to be rotated based on the current date.
        """
        if time.strftime('%Y-%m-%d') != self.day:
            self.day = time.strftime('%Y-%m-%d')
            self.setup()
            self.delete_expired_log()
        
    def shutdown(self):
        """
        Properly close and remove all handlers from the logger.
        """
        if not hasattr(self, 'logger'):
            return
        for handler in self.logger.handlers:
            print(handler, flush=True)
            handler.close()
            self.logger.removeHandler(handler)


import time
import json
import logging
import threading
import numpy as np

from nxva.v11 import Detector
from nxva.streaming import MultiStreaming
from nxva.va.shm_utils import SHMHandler
from collections import deque
from flask import Flask, request, jsonify

app = Flask(__name__)

class DetectionServer:
    def __init__(self, max_detection_num, model_list, allowed_class_names, all_cameras_ids, interval=0, namespace=None):
        '''
        Args:
            max_detection_num: int
                max number of detection in a frame
            model_list: list
                list of model names
                model_list = ['./configs/detection.yaml', ...]
            allowed_class_names: list
                list of class names which the task needs
                allowed_class_names = ['people', 'car', ...]
            all_cameras_ids: list
                append all cameras.yaml where in cameras's folder
            interval: int
                interval time for detection, default is 0
            namespace: str
                namespace for logger, default is None

        Attributes:
            max_detection_num: int
                max number of detection in a frame
            all_cameras_ids: list
                append all cameras.yaml where in cameras's folder
            stream: MultiStreaming
                streaming
            detector: DetectorBuilder
                detector
            model_dict: dict
                run models by model_name
                model_dict[model_name] = model
            class_name_dict: dict
                send class names to client after label shift
                class_name_dict = {0: 'people', ...}
            label_shift: dict
                merge all labels
                label_shift[model_name] = shift
            task_list: list
                list of task names
                task_list = [task1, task2, ...]
            mix_name_classes: dict
                mix_name_classes = {class_name: class_id, ...}
            frames_results_queue
                frames_results_queue = queue.Queue()
            queue_lock: threading.Lock
                lock for frames_results_queue
        '''
        self.logger = logging.getLogger(namespace) if namespace else logging.getLogger()

        self.app = Flask(__name__)
        self.setup_routes()

        self.max_detection_num     = max_detection_num
        self.interval              = interval

        self.all_cameras_ids       = all_cameras_ids
        self.cameras_num           = len(all_cameras_ids) # camera數量

        # streaming
        self.stream = MultiStreaming(config='./configs/streaming.yaml', 
                                    reconnect_interval=30,
                                    verbose=True
                                    )
        self.stream.run()

        # model yaml
        self.mix_class_names  = {}
        self.model_dict       = {}
        self.label_shift      = {}
        shift                 = 0
        
        model_name = 0
        for cfg in model_list:
            model = Detector(cfg)
            model.classes = []
            model.max_det = max_detection_num
            model.name_classes = {name: class_id for class_id, name in model.class_names.items()}
            self.model_dict[model_name] = model
            model_name += 1

        for cls_name in allowed_class_names:
            for model_name, model in reversed(self.model_dict.items()):
                class_id = model.name_classes.get(cls_name)
                if class_id is not None:
                    model.classes.append(class_id)

        shift = 0
        for model_name, model in self.model_dict.items():
            for class_id, cls_name in model.class_names.items():
                self.mix_class_names[class_id + shift] = cls_name
            self.label_shift[model_name] = shift
            shift = len(self.mix_class_names)

        self.mix_name_classes     = {name: class_id for class_id, name in self.mix_class_names.items()}
        
        # task for debug
        self.task_list            = []

        # queue
        self.frames_results_queue = deque(maxlen=1)

        # lock
        self.queue_lock           = threading.Lock()

        # SHMHandler
        self.shm_handler          = SHMHandler(self.stream.w, self.stream.h, self.cameras_num, max_detection_num, namespace=namespace)

        # shared memory
        self.shm_handler.build_shared_memory()

        # init control block
        self.shm_handler.write_control_info(slot_idx=0)

        # stop event
        self.stop_event           = threading.Event()
        self.interval_stop_event  = threading.Event()

        threading.Thread(target=self.api_run, daemon=True).start()
    
    
    def setup_routes(self):
        @app.route('/api/registration_server', methods=['POST'])
        def registration_server():
            '''
            Registration server.
            Listen to the client and send the shared memory info.
            body:
                {
                "camera_ids"    : camera_ids,
                "class_names"   : class_names,
                }
            '''
            self.logger.info('Waiting for connection...')

            message       = request.get_json()
            cameras_ids   = message['camera_ids']
            class_names   = message['class_names']

            task_count = len(self.task_list)
            name = f"task_{task_count}"
            
            # find camera_ids in all_cameras_ids
            camera_indices_list = [
                self.all_cameras_ids.index(task_camera) 
                for task_camera in cameras_ids
                if task_camera in self.all_cameras_ids
            ]

            # find class_id in class_names
            class_name_dict = {
                self.mix_name_classes[name]: name
                for name in class_names
                if name in self.mix_name_classes
            }

            self.task_list.append(name)
            self.logger.info(f"Register task: {self.task_list}")

            respone = {
                'shm_name':                 "frames_results_shm",
                'total_cameras_num':        self.cameras_num,
                'camera_ids':               camera_indices_list,
                'width':                    self.stream.w,
                'height':                   self.stream.h,
                'class_name_dict':          class_name_dict,
                'max_detection_num':        self.max_detection_num,
            }

            return jsonify(respone), 200

        @app.route('/api/watchdog_server', methods=['POST'])
        def watchdog_server():
            '''
            Check if the cameras is running.
            body:
                {
                "device_id": "CAMERA-201-1"
                }
            '''
            data      = request.get_json()
            device_id = data['device_id']

            camera_index  = self.all_cameras_ids.index(device_id)
            camera_status = self.stream.status[camera_index]

            server_available    = []
            server_unavailable  = []
            if camera_status:
                server_available.append(device_id)
            else:
                server_unavailable.append(device_id)

            data = {
                'server_available': server_available,
                'server_unavailable': server_unavailable
                }

            return jsonify(data), 200
        
    def api_run(self):
        app.run(host='0.0.0.0', port=9453)
    
    def run(self):
        '''
        Run get_frames_results_queue threading.
        Get frames and results from stream, and put them into queue.
        '''
        self.stop_event.clear()
        num_frames = self.stream.num
        threading.Thread(target=self._write_frames_results_shm, daemon=True).start()
        next_time = time.perf_counter() + self.interval
        while not self.stop_event.is_set():
            merged_results = [np.empty((0, 6), dtype=float) for _ in range(num_frames)]
            frames, status = self.stream.get_frames(status=True)

            for model_name, model in self.model_dict.items():
                results = model(frames)

                # Shift class_id and merge results
                for frame_idx, result in enumerate(results):
                    if result.shape[1] >= 6:
                        result[:, 5] += self.label_shift[model_name]
                    merged_results[frame_idx % num_frames] = np.vstack((merged_results[frame_idx % num_frames], result))

            with self.queue_lock:
                self.frames_results_queue.append((frames, merged_results))

            # Calculate the next time to send the next frame
            next_time += self.interval
            delay = next_time - time.perf_counter()
            if delay > 0:
                self.interval_stop_event.wait(delay)
            else:
                time.sleep(0.001)
            
            time.sleep(0.01)
    
    def _write_frames_results_shm(self):
        '''
        If queue is not empty, get frames and results from queue, and put them into frames_buffer and results_buffer.

        1) Get frames and results from queue
        2) Write frames and results to shared memory
        3) Write control info to shared memory
        '''

        frame_counter = 0
        while not self.stop_event.is_set():
            if len(self.frames_results_queue) > 0:
                with self.queue_lock:
                    frames, results = self.frames_results_queue.popleft()
                
                for cam_id in range(self.cameras_num):

                    # write frames and metas to shared memory
                    old_slot_idx = self.shm_handler.read_control_info()

                    # buffer index (0->1 or 1->0)
                    next_slot_idx = 1 - old_slot_idx

                    self.shm_handler.write_frame_and_bbox(cam_id, next_slot_idx, frames[cam_id], results[cam_id])

                # frame_counter limit, results
                if frame_counter <= self.stream.h*self.stream.w:
                    frame_counter += 1
                else:
                    frame_counter = 0

                meta_dict = {
                    "camera_id": cam_id,
                    "frame_counter": frame_counter,
                    "timestamp": time.time()
                }
                meta_str = json.dumps(meta_dict)
                
                self.shm_handler.write_control_info(next_slot_idx)
                self.shm_handler.write_meta_info(meta_str)

                    
            time.sleep(0.001)

    def stop(self):
        '''
        Stop the detection server.
        '''
        self.stop_event.set()
        self.logger.info('Detection server stopped.')


if __name__ == "__main__":
    from utilities import load_configs

    task_list, hyper_config, streaming_config, cameras_dict, cameras_configs_dict, all_cameras_ids = load_configs()

    max_detection_num = hyper_config['shared_memory']['max_detection_num']
    model_list        = hyper_config['model']['model_list']
    class_names  = [config['class'] for _, config in hyper_config.items() if 'class' in config]
    allowed_class_names = list(set(name for class_name in class_names for name in class_name))

    dection_server = DetectionServer(max_detection_num, model_list, allowed_class_names, all_cameras_ids)
    dection_server.run()


from dataclasses import dataclass
from typing import Union, List, Optional
from ultralytics import YOLO
import yaml
import torch
import numpy as np
from pathlib import Path
from PIL import Image
import warnings

@dataclass(frozen=True)
class ClassifierOptions:
    weights: str = './weights/yolo11s-cls.pt'
    size: Union[int, tuple] = 640
    device: str = None
    class_names: dict = None
    fp16: bool = False
    verbose: bool = False
    batch: int = 1
    visualze: bool = False
    augment: bool = False
    embed: Optional[List[int]] = None
    project: str = None
    name: str = None

class Classifier():
    """
    A class to customize YOLOv11 classification output format based on a YAML configuration or an object.

    Attributes:
        weights (str): Path to the model weights file.
        size (int or tuple): Defines the image size for inference. Can be a single integer for square resizing
            or a (height, width) tuple.
        device (str): Specifies the device for inference.
        class_names (dict): Dictionary mapping class indices to class names.
        fp16 (bool): Whether to use half precision (FP16) for inference.
        verbose (bool): If True, enables verbose output during the model's initialization and subsequent operations.
        batch (int): Specifies the batch size for inference.
        visualize (bool): Activates visualization of model features during inference.
        augment (bool): Enables test-time augmentation (TTA) for predictions, potentially improving detection
            robustness at the cost of inference speed.
        embed (list): Specifies the layers from which to extract feature vectors or embeddings.
        project (str): Name of the project directory where prediction outputs are saved if save is enabled.
        name (str): Name of the prediction run. Used for creating a subdirectory within the project folder.
    """
    def __init__(self, config: Union[str, ClassifierOptions]):
        """
        Initialize the Classifier class.
        
        Args:
            config (Union[str, ClassifierOptions]): A YAML configuration file path or a 'ClassifierOptions' object.
        Raises:
            ValueError: If the config isn't a YAML file or 'ClassifierOptions' object.

        Example:
            1. YAML file
                >>> model = Classifier("config.yaml")
                >>> results = model(images)
                >>> print(results)
            2. object
                >>> options = ClassifierOptions(verbose=True....)
                >>> model = Classifier(options)
                >>> results = model(images)
                >>> print(results)
        """
        # Load config from a YAML file or a ClassifierOptions object
        if isinstance(config, str) and config.endswith(('yaml', 'yml')):
            with open(config, 'r') as f:
                self.config = yaml.safe_load(f)

        elif isinstance(config, ClassifierOptions):
            self.config = {k: v for k, v in config.__dict__.items() if not k.startswith("__")}

        else:
            raise ValueError("Config must be a YAML file or object.")

        # Initialize attributes with values from the configuration
        self.weights = self.config['weights']
        self.size = self.config.get('size', 640)
        self.device = self.config.get('device', None)
        self.class_names = self.config.get('class_names', None)
        self.batch = self.config.get('batch', 1)
        self.visualize = self.config.get('visualize', False)
        self.augment = self.config.get('augment', False)
        self.embed = self.config.get('embed', None)
        self.fp16 = self.config.get('fp16', False)
        self.verbose = self.config.get('verbose', False)
        self.project = self.config.get('project', None)
        self.name = self.config.get('name', None)
        
        # Load model
        self.model = self._load_model()

        # Initialize class names
        self._initialize_class_names()
    
    def _load_model(self):
        """Load the classification model based on YOLO."""
        model = YOLO(model = self.weights, task='classify')
        return model
    
    def _initialize_class_names(self):
        """Initialize the class names."""
        # Initialize class_names if not already set
        if not self.class_names:
            if hasattr(self.model, 'names') and self.model.names:
                self.class_names = self.model.names
            else:
                warnings.warn("No 'class_names' found in the configuration.", UserWarning)
    
    def __call__(self, images, top_k=1, *args, **kwargs):
        """
        Performs inference on the given image source and customizes the output format.

        Args:
            images (str | Path | int | PIL.Image | np.ndarray | torch.Tensor | List | Tuple): The input image(s)
                to performs inference on. Accepts various types including file paths, URLs, PIL images, numpy arrays,
                and torch tensors.
            **kwargs: Additional keyword arguments passed to the model inference.

        Return:
            (List): A list of customized output results, where each result contains the top-k predictions for the
                corresponding input.
        """
        # Check images
        if isinstance(images, (np.ndarray, torch.Tensor)):
            if images.shape[0] == 0:
                print("The input image is empty", flush=True)
                return []
        elif isinstance(images, (list, tuple)):
            if not len(images):
                print("The input images list or tuple is empty", flush=True)
                return []
        # elif isinstance(images, (str, int, Path)):
        #     if not images:
        #         print("The source path, URL, or camera ID is empty", flush=True)
        #         return []
        elif isinstance(images, Image.Image):
            if images.size == (0, 0):
                print("The input PIL image is empty", flush=True)
                return []

        # Performs inference
        results = self.model(
            source = images,
            imgsz = self.size,
            device = self.device,
            batch = self.batch,
            visualize = self.visualize,
            augment = self.augment,
            embed = self.embed,
            half = self.fp16,
            verbose = self.verbose,
            project = self.project,
            name = self.name,
            **kwargs
        )

        # Customizes the output format
        customed_results = []
        for result in results:
            all_class_ids = result.probs.data
            sorted_class_ids = (-all_class_ids).argsort(0).tolist()
            confidences = all_class_ids[sorted_class_ids].cpu().numpy()
            
            if top_k < 0:
                selected_indices = range(len(sorted_class_ids)) # Displays all rankings
            else:
                selected_indices = range(min(top_k, len(sorted_class_ids))) # Displays the top_k rankings.

            formatted_result = {sorted_class_ids[i] : confidences[i] for i in selected_indices}
            customed_results.append(formatted_result)
        return customed_results
            
    def destroy(self):
        """
        Clean up resources.

        Deletes the model instance to release memory and ensure proper resource cleanup.
        """
        if self.model:
            del self.model


import cv2
import copy
import math
import numpy as np

color_list = [
    [0, 0, 255],     
    [0, 255, 0],     
    [255, 0, 0],     
    [0, 255, 255],   
    [255, 255, 0],   
    [255, 0, 255],   
    [0, 165, 255],   
    [203, 192, 255], 
    [139, 0, 0],     
    [255, 228, 225], 
    [0, 0, 139],     
    [144, 238, 144], 
    [0, 128, 128],   
    [128, 0, 128],   
    [130, 0, 75],    
    [19, 69, 139],   
    [220, 220, 220], 
    [0, 215, 255],   
    [192, 192, 192], 
    [128, 128, 128], 
    [211, 211, 211], 
    [169, 169, 169], 
    [255, 191, 0],   
    [0, 255, 224],   
    [147, 112, 219], 
    [0, 100, 0],     
    [127, 255, 170], 
    [255, 105, 180],
    [0, 0, 0],       
    [255, 255, 255],  
]

def frame_cut(frame, bbox):
    """
    params:
        frame: np.ndarray
        bbox: list or 1-d array, [x1, y1, x2, y2]
    return:
        np.ndarray
    """
    cut_bbox = copy.deepcopy(bbox)
    if bbox[0] > bbox[2]:
        cut_bbox[0] = bbox[2]
        cut_bbox[2] = bbox[0]
    if bbox[1] > bbox[3]:
        cut_bbox[1] = bbox[3]
        cut_bbox[3] = bbox[1]
    return frame[int(cut_bbox[1]):int(cut_bbox[3]), int(cut_bbox[0]):int(cut_bbox[2])]


def draw_text(img, 
              text,
              font=cv2.FONT_HERSHEY_SIMPLEX,
              font_scale=0.5,
              font_thickness=1,
              text_color=(0, 0, 255),
              text_color_bg=(255, 255, 255),
              pos='tl',
              axis=(0, 0)):
    """
    This function is used to draw text on the image.
    The axis is the top-left corner of the text if the pos is 'tl'.
    params:
        img: np.ndarray
        text: str
        font: int
        font_scale: float
        font_thickness: int
        text_color: tuple
        text_color_bg: tuple
        pos: str, 'tl', 'tr', 'bl', 'br'
        axis: tuple, (x, y)
    return:
        tuple, (text_w, text_h)
    """
    axis = (int(axis[0]), int(axis[1]))
    text_size, _ = cv2.getTextSize(text, font, font_scale, font_thickness)
    text_w, text_h = text_size
    try:
        if pos == 'bl':
            if text_color_bg:
                cv2.rectangle(img, axis, (axis[0]+text_w, axis[1]-text_h*2), text_color_bg, -1)
            cv2.putText(img, text, (axis[0], int(axis[1] - text_h/2)), font, font_scale, text_color, font_thickness, cv2.LINE_AA)
        elif pos == 'tl':
            if text_color_bg:
                cv2.rectangle(img, axis, (axis[0]+text_w, axis[1]+text_h*2), text_color_bg, thickness=-1)
            cv2.putText(img, text, (axis[0], int(axis[1] + text_h*3/2)), font, font_scale, text_color, font_thickness, cv2.LINE_AA)
        elif pos == 'tr':
            if text_color_bg:
                cv2.rectangle(img, axis, (axis[0]-text_w, axis[1]+text_h*2), text_color_bg, thickness=-1)
            cv2.putText(img, text, (axis[0]-text_w, int(axis[1] + text_h*3/2)), font, font_scale, text_color, font_thickness, cv2.LINE_AA)
        elif pos == 'br':
            if text_color_bg:
                cv2.rectangle(img, axis, (axis[0]-text_w, axis[1]-text_h*2), text_color_bg, thickness=-1)
            cv2.putText(img, text, (axis[0]-text_w, int(axis[1] - text_h/2)), font, font_scale, text_color, font_thickness, cv2.LINE_AA)
    except:
        print('position and axis are wrong setting')
        
    return text_w, text_h * 2
        
    
def draw_bbox(img, 
              result, 
              idxs=None, 
              classes=None, 
              str_list=[],
              score=False, 
              track_id=False,
              font=cv2.FONT_HERSHEY_SIMPLEX,
              font_scale=0.5,
              font_thickness=1,
              color=None):
    """
    Draw bounding boxes on the image according to the detection result.
    img will be modified in place.
    params:
        img: np.ndarray
        result: np.ndarray, [x1, y1, x2, y2, score, class, uid]
        classes: list
        score: bool
        str_list: list
        font: int
        font_scale: float
        font_thickness: int
        idxs: list
        color: tuple
    """
    if isinstance(result, np.ndarray):
        if result.ndim == 1:
            result = np.expand_dims(result, 0)
        for j in range(result.shape[0]):
            if idxs is None or result[j, 5] in idxs:
                _color = color if color else color_list[int(result[j, 5]) % len(color_list)]    
                    
                cv2.rectangle(img, (int(result[j, 0]), int(result[j, 1])), (int(result[j, 2]), int(result[j, 3])), _color, 2)
                
                if track_id:
                    bias_w, _ = draw_text(img, str(int(result[j, 6])),
                                          text_color=(255, 255, 255), text_color_bg=_color, 
                                          pos='bl', axis=(int(result[j, 0]), int(result[j, 1])))
                else:
                    bias_w = 0
                
                if classes:
                    text = str(classes[int(result[j, 5])])
                    if score:
                        text += ': ' + str(round(result[j, 4], 2))
                    cv2.putText(img, text, (int(result[j, 0]) + bias_w, int(result[j, 1])-5), font, font_scale, _color, font_thickness, cv2.LINE_AA)
                elif str_list:
                    cv2.putText(img, str_list[j], (int(result[j, 0]) + bias_w, int(result[j, 1])-5), font, font_scale, _color, font_thickness, cv2.LINE_AA)


def images_automerge(images, size):
    """
    Merge images into one image.
    params:
        images: list
        size: tuple
    return:
        np.ndarray
    """
    num = (math.ceil(len(images) ** 0.5))
    if isinstance(size, int):
        size = (size, size)
    w, h = size
    
    result = np.zeros((h, w, 3), dtype='uint8')
    w = int(w / num)
    h = int(h / num)
    for i, img in enumerate(images):
        r = int(i / num)
        c = int(i % num)
        result[r*h:(r+1)*h, c*w:(c+1)*w, :] = cv2.resize(img, (w, h))
        
    return result

def draw_mask_on_img(img, mask, reverse=False, alpha=0.5, color=[0, 0, 255]):
    img_uint8 = img.astype(np.uint8)
    overlay = np.zeros_like(img_uint8)
    overlay[:, :] = color
    if reverse:
        mask = ~(mask.astype(bool))
    mask_uint8 = (mask * 255).astype(np.uint8)  # ensure the mask is in uint8 not bool
    masked_overlay = cv2.multiply(overlay, np.stack([mask_uint8]*3, axis=-1))
    result = cv2.addWeighted(img_uint8, 1, masked_overlay, alpha, 0)
    
    return result


import time
import torch

# can only decorate on instance method
def timer():
    def decorator(func):
        def wrapper(*args, **kwargs):
            self = args[0]  # 確定其中之一是 `self`
            if hasattr(self, 'verbose') and self.verbose:
                torch.cuda.synchronize()
                start_time = time.time()
                result = func(*args, **kwargs)
                torch.cuda.synchronize()
                end_time = time.time()
                print(f"{func.__name__}: {(end_time - start_time)*1000:.1f} ms")
            else:
                result = func(*args, **kwargs)
            return result
        return wrapper
    return decorator


import numpy as np
import torch

def xyxy2xywh(x):
    """
    Convert bounding box coordinates from (x1, y1, x2, y2) format to (x, y, width, height) format.

    Args:
        x (np.ndarray) or (torch.Tensor): The input bounding box coordinates in (x1, y1, x2, y2) format.
    Returns:
       y (np.ndarray) or (torch.Tensor): The bounding box coordinates in (x, y, width, height) format.
    """
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[..., 0] = (x[..., 0] + x[..., 2]) / 2  # x center
    y[..., 1] = (x[..., 1] + x[..., 3]) / 2  # y center
    y[..., 2] = x[..., 2] - x[..., 0]  # width
    y[..., 3] = x[..., 3] - x[..., 1]  # height
    return y



def nn_cosine_distance(x, y):
    """ Helper function for nearest neighbor distance metric (cosine).
    Parameters
    ----------
    x : ndarray
        A matrix of N row-vectors (sample points).
    y : ndarray
        A matrix of M row-vectors (query points).
    Returns
    -------
    ndarray
        A vector of length M that contains for each entry in `y` the
        smallest cosine distance to a sample in `x`.
    """
    x_ = torch.from_numpy(np.asarray(x, dtype=np.float64))
    y_ = torch.from_numpy(np.asarray(y, dtype=np.float64))
    # x_ = torch.from_numpy(np.asarray(x))
    # y_ = torch.from_numpy(np.asarray(y))
    distances = _cosine_distance(x_, y_)
    distances = distances
    # return distances.min(axis=0)
    return distances.max(axis=0) #改成愈大愈好

def _cosine_distance(a, b, data_is_normalized=False): #計算特徵距離（愈大愈好）
    """Compute pair-wise cosine distance between points in `a` and `b`.
    Parameters
    ----------
    a : array_like
        An NxM matrix of N samples of dimensionality M.
    b : array_like
        An LxM matrix of L samples of dimensionality M.
    data_is_normalized : Optional[bool]
        If True, assumes rows in a and b are unit length vectors.
        Otherwise, a and b are explicitly normalized to lenght 1.
    Returns
    -------
    ndarray
        Returns a matrix of size len(a), len(b) such that eleement (i, j)
        contains the squared distance between `a[i]` and `b[j]`.
    """
    if not data_is_normalized:
        a = np.asarray(a) / np.linalg.norm(a, axis=1, keepdims=True)
        b = np.asarray(b) / np.linalg.norm(b, axis=1, keepdims=True)
    # return 1 - np.dot(a, b.T)
    return np.dot(a, b.T) #改成愈大愈好

def k_previous_obs(observations, cur_age, k):
    if len(observations) == 0:
        return [-1, -1, -1, -1, -1]
    for i in range(k):
        dt = k - i
        if cur_age - dt in observations:
            return observations[cur_age-dt]
    max_age = max(observations.keys())
    return observations[max_age]

def get_speed_direction(pre_time, cur_time, pre_box, pre_n_box, cur_box, pos_lambda, direction_type='vector'):


    pre_n_cx = (1 - pos_lambda[0]) * pre_n_box[0] + pos_lambda[0] * pre_n_box[2]
    pre_n_cy = (1 - pos_lambda[1]) * pre_n_box[1] + pos_lambda[1] * pre_n_box[3]

    pre_cx = (1 - pos_lambda[0]) * pre_box[0] + pos_lambda[0] * pre_box[2]
    pre_cy = (1 - pos_lambda[1]) * pre_box[1] + pos_lambda[1] * pre_box[3]

    cur_cx = (1 - pos_lambda[0]) * cur_box[0] + pos_lambda[0] * cur_box[2]
    cur_cy = (1 - pos_lambda[1]) * cur_box[1] + pos_lambda[1] * cur_box[3]


    delta_1_x = cur_cx - pre_cx
    delta_1_y = cur_cy - pre_cy
    
    delta_n_x = cur_cx - pre_n_cx
    delta_n_y = cur_cy - pre_n_cy
    
    # speed ( ||delta_x,delta_y|| / ( tiem.time() - pre_time) )
    delta_mov = np.sqrt(delta_1_x ** 2 + delta_1_y ** 2) #+ 1e-6
    speed = delta_mov / (cur_time - pre_time)

    if direction_type == 'angle':
        # direction (delta_y/delta_x) 水平3點鐘方向=0度, 逆時針0~360度
        velocity_direction = np.degrees(np.arctan2(delta_n_y, delta_n_x))
        if velocity_direction < 0 :
            velocity_direction +=360
        return speed, velocity_direction

    elif direction_type == 'vector':
        vector = (delta_n_x, delta_n_y)
        return speed, vector
    
    elif direction_type == 'radian':
        radian = np.arctan2(delta_n_y, delta_n_x)
        return speed, radian
    
    else:
        # TODO
        print('we don"t have this direction type')
        return None






import os
import cv2
import sys
import yaml
import numpy as np
import torch
import torch.nn.functional as F
import torchvision.transforms as T
from nxva.utilities import timer
from typing import Union
from dataclasses import dataclass
import warnings

IMAGENET_MEAN = 0.485, 0.456, 0.406  # RGB mean
IMAGENET_STD = 0.229, 0.224, 0.225  # RGB standard deviation

class CenterCrop:
    # YOLOv5 CenterCrop class for image preprocessing, i.e. T.Compose([CenterCrop(size), ToTensor()])
    def __init__(self, size=640):
        super().__init__()
        self.h, self.w = (size, size) if isinstance(size, int) else size

    def __call__(self, im):  # im = np.array HWC
        imh, imw = im.shape[:2]
        m = min(imh, imw)  # min dimension
        top, left = (imh - m) // 2, (imw - m) // 2
        return cv2.resize(im[top:top + m, left:left + m], (self.w, self.h), interpolation=cv2.INTER_LINEAR)
    
    
class ToTensor:
    # YOLOv5 ToTensor class for image preprocessing, i.e. T.Compose([LetterBox(size), ToTensor()])
    def __init__(self, half=False):
        super().__init__()
        self.half = half

    def __call__(self, im):  # im = np.array HWC in BGR order
        im = np.ascontiguousarray(im.transpose((2, 0, 1))[::-1])  # HWC to CHW -> BGR to RGB -> contiguous
        im = torch.from_numpy(im)  # to torch
        im = im.half() if self.half else im.float()  # uint8 to fp16/32
        im /= 255.0  # 0-255 to 0.0-1.0
        return im
    

def classify_transforms(size=640):
    # Transforms to apply if albumentations not installed
    assert isinstance(size, int), f'ERROR: classify_transforms size {size} must be integer, not (list, tuple)'
    # T.Compose([T.ToTensor(), T.Resize(size), T.CenterCrop(size), T.Normalize(IMAGENET_MEAN, IMAGENET_STD)])
    return T.Compose([CenterCrop(size), ToTensor(), T.Normalize(IMAGENET_MEAN, IMAGENET_STD)])

class BaseClassifier:
    """
    BaseClassifier.

    A base class for creating classifier.

    Attributes:
        model_path (str): Path to the model file.
        device (str): "cuda" or "cpu".
        size (int): The size of the input image.
        fp16 (bool): Flag for using FP16 precision during inference.
        transforms (callable): Transformation function for preprocessing input images.
        class_names (list): List of class names for the model's outputs.
    """
    def __init__(self, model_path, device, size, class_names=None, fp16=False, verbose=False):
        """
        Initialize the BaseClassifier class.

        Args:
            model_path (str): Path to the model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            verbose (bool): Whether to output the timer.
        """
        assert model_path, 'ERROR: model_path required.'
        self.model_path = model_path
        self.device = device
        self.size = size
        self.fp16 = fp16
        self.class_names = class_names
        self.transforms = classify_transforms(self.size)
        self.model = self.load_model()

        if self.class_names is None:
            if hasattr(self.model, 'module') and self.model.class_names:
                self.class_names = self.model.module.names
            elif hasattr(self.model, 'names') and self.model.names:
                self.class_names = self.model.names
            else:
                warnings.warn("No 'class_names' found in the configuration.", UserWarning)
        
    def load_model(self):
        raise NotImplementedError("Subclasses should overwrite this method to load the model!")

    @timer()
    def preprocess(self, imgs):
        """Prepares input image before inference."""
        if isinstance(imgs, list):
            imgs = [self.transforms(img) for img in imgs]
            imgs = torch.stack(imgs)
        else:
            imgs = self.transforms(imgs)
            imgs = imgs.unsqueeze(0)
        
        imgs = imgs.to(self.device)

        if self.fp16:
            imgs = imgs.half()
        return imgs
    
    @timer()
    def post_process(self, results, top_k=1):
        """Post-processes predictions for an image and returns them."""
        ret = []  
        top_k = min(top_k, len(self.class_names))
        for result in results:
            if top_k < 0:
                rank = np.argsort(result)[::-1] # Displays all rankings
            else:
                rank = np.argsort(result)[::-1][:top_k] # Displays the top_k rankings.

            processed_result = {i: result[i] for i in rank}
                
            ret.append(processed_result)
            
        return ret

    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the specified model."""
        raise NotImplementedError("Subclasses should overwrite this method to run the classification!")
    
    def destroy(self):
        pass
    
    def __call__(self, imgs, top_k=1):
        """Performs inference on the given image source."""
        pre_imgs = self.preprocess(imgs) 
        if not self.needs_tensor():
            pre_imgs = pre_imgs.cpu().numpy()

        results = self.infer(pre_imgs)
        ret = self.post_process(results, top_k = top_k)
        return ret
        
    def needs_tensor(self):
        return True

class TorchClassifier(BaseClassifier):
    """
    TorchClassifier.

    A class for loading and running classification models in PyTorch format.
    """
    def __init__(self, model_path, device, size, class_names=None, fp16=False, verbose=False):
        """
        Initialize the TorchClassifier class.

        Args:
            model_path (str): Path to the PyTorch model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            verbose (bool): Whether to output the timer.
        """
        super().__init__(model_path, device, size, class_names, fp16, verbose)

    def load_model(self):
        """Load PyTorch model."""
        current_file_path = os.path.abspath(__file__)
        current_dir = os.path.dirname(current_file_path)
        sys.path.insert(0, os.path.join(current_dir, 'yolov5'))
        from models.experimental import attempt_load
        sys.path.remove(os.path.join(current_dir, 'yolov5'))

        model = attempt_load(self.model_path, device=self.device)
        model = model.half().eval() if self.fp16 else model.float().eval()
        return model
    
    @timer()
    def infer(self, imgs):  
        """Runs inference on a given image using the PyTorch model."""
        with torch.no_grad():
            results = self.model(imgs)
            results = F.softmax(results, dim=1)  # probabilities
            results = results.cpu().numpy()
        return results
    
class ONNXClassifier(BaseClassifier):
    """
    ONNXClassifier.

    A class for loading and running classification models in ONNX format.

    Attributes:
        size (tuple): The size of the model input, typically (width, height).
    """
    def __init__(self, model_path, device, size, class_names=None, fp16=False, verbose=False):
        """
        Initialize the ONNXClassifier class.

        Args:
            model_path (str): Path to the ONNX model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            verbose (bool): Whether to output the timer.
        """
        super().__init__(model_path, device, size, class_names, fp16, verbose)
        self.size = self.determine_input_size()
        assert self.size == (size, size) ,"The input size of the model is not the same as the size in the config file."

    def load_model(self):
        """Load ONNX model."""
        import onnxruntime as ort
        if self.device == 'cuda':
            print("Using CUDA")
            providers = [("CUDAExecutionProvider", {"device_id": 0})]
        else:
            print("Using CPU")
            providers = ["CPUExecutionProvider"]

        return ort.InferenceSession(self.model_path, providers=providers)
    
    def determine_input_size(self):
        """Determines the input size expected by the model."""
        input_shape = self.model.get_inputs()[0].shape
        assert len(input_shape) == 4, "Input shape is not 4D" 
        assert input_shape[2] is not None and input_shape[3] is not None, "Input height/width is None"
        return (input_shape[2], input_shape[3])
    
    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the ONNX model."""
        results = self.model.run(None, {self.model.get_inputs()[0].name: imgs})  # Perform inference
        results = F.softmax(torch.tensor(results[0]), dim=1)  # Convert result to tensor and apply softmax
        results = results.cpu().numpy()  # Move to CPU and convert to NumPy
        return results
    
    def needs_tensor(self):
        return False

class TensorRTClassifier(BaseClassifier):
    """
    TensorRTClassifier.

    A class for loading and running classification models in TensorRT format.

    Attributes:
        trt (module): TensorRT module.
        cuda (module): PyCUDA module.
        engine (TensorRT.ICudaEngine): TensorRT engine.
        context (TensorRT.IExecutionContext): TensorRT execution context.
        size (tuple): The size of the input image.
        fixed_batch (int): The fixed batch size of the model.
        bindings (list): List of buffer bindings. Pointer to the input and output buffers.
        inputs (list): List of input buffers.  
        outputs (list): List of output buffers.
        stream (pycuda.driver.Stream): CUDA stream. Used for asynchronous memory transfers.
    """
    def __init__(self, model_path, device, size, class_names=None, fp16=False, verbose=False):
        """
        Initialize the TensorRTClassifier class.

        Args:
            model_path (str): Path to the PyTorch model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            verbose (bool): Whether to output the timer.
        """
        assert device == "cuda", "TensorRT only supports CUDA."
        self.trt, self.cuda = self._import_dependencies()
        super().__init__(model_path, device, size, class_names, fp16, verbose)
        self.engine = self.model
        self.context = self.engine.create_execution_context()  # create a context
        self.size, self.fixed_batch = self.get_input_size()
        assert self.size == (size, size), "The input size of the model is not the same as the size in the config file."
        self.bindings, self.inputs, self.outputs, self.stream = self.prepare_buffers()

    def _import_dependencies(self):
        import tensorrt as trt
        import pycuda.driver as cuda
        import pycuda.autoinit  # Ensures initialization of CUDA

        return trt, cuda
    
    def load_model(self):
        """Load TensorRT model."""
        logger = self.trt.Logger(self.trt.Logger.INFO)
        with open(self.model_path, "rb") as f, self.trt.Runtime(logger) as runtime:
            return runtime.deserialize_cuda_engine(f.read())
    
    def get_input_size(self):
        """Determines the input size expected by the model."""
        for binding_index in range(self.engine.num_bindings):
            if self.engine.binding_is_input(binding_index):
                shape = self.engine.get_binding_shape(binding_index)
                return (shape[2], shape[3]), shape[0]
    
    def prepare_buffers(self):
        """Prepare input and output buffers for TensorRT inference."""
        bindings = [] 
        stream = self.cuda.Stream() 
        inputs, outputs = [], [] 

        for binding_index in range(self.engine.num_bindings): 
            binding_shape = self.engine.get_binding_shape(binding_index) #(B. C, H, W)
            if self.fp16 and self.engine.get_binding_dtype(binding_index) == self.trt.DataType.FLOAT: 
                binding_dtype = np.float16 
            else:
                binding_dtype = self.trt.nptype(self.engine.get_binding_dtype(binding_index)) 
            binding_size = self.trt.volume(binding_shape) * np.dtype(binding_dtype).itemsize 
            device_mem = self.cuda.mem_alloc(binding_size) 
            bindings.append(int(device_mem)) 
            
            if self.engine.binding_is_input(binding_index): 
                host_mem = np.zeros(binding_shape, dtype=binding_dtype)
                inputs.append({'host': host_mem, 'device': device_mem})
            else:
                host_mem = np.zeros(binding_shape, dtype=binding_dtype)
                outputs.append({'host': host_mem, 'device': device_mem})

        return bindings, inputs, outputs, stream
    
    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the TensorRT model."""
        try:
            batch_size = imgs.shape[0]
            if batch_size > self.fixed_batch:
                raise ValueError(f"Input batch size {batch_size} exceeds the fixed batch size {self.fixed_batch}.")
            if batch_size < self.fixed_batch:
                padding = np.zeros((self.fixed_batch - batch_size, *imgs.shape[1:]), dtype=imgs.dtype)
                imgs = np.concatenate([imgs, padding], axis=0) 

            # Transfer input data to the GPU
            self.cuda.memcpy_htod_async(self.inputs[0]['device'], imgs, self.stream)

            # Run inference
            self.context.execute_async_v2(bindings=self.bindings, stream_handle=self.stream.handle)

            # Transfer predictions back from the GPU
            for output in self.outputs:
                self.cuda.memcpy_dtoh_async(output['host'], output['device'], self.stream)

            self.stream.synchronize()
            results = self.outputs[0]['host'][0:batch_size]
            results = F.softmax(torch.tensor(results), dim=1) # Convert result to tensor and apply softmax
            results = results.cpu().numpy()  # Move to CPU and convert to NumPy
        except self.cuda.LogicError as e:
            print(f"CUDA logic error: {str(e)}")
        except self.cuda.MemoryError as e:
            print(f"CUDA memory error: {str(e)}")
        except Exception as e:
            print(f"General error during CUDA operations: {str(e)}")

        return results
    
    def needs_tensor(self):
        return False
    
class Torch2trtClassifier(BaseClassifier):
    """
    Torch2trtClassifier.

    A class for loading and running classification models in trt format.
    """
    def __init__(self, model_path, device, size, class_names=None, fp16=False, verbose=False):
        """
        Initialize the Torch2trtClassifier class.

        Args:
            model_path (str): Path to the trt model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            verbose (bool): Whether to output the timer.
        """
        super().__init__(model_path, device, size, class_names, fp16, verbose)
        
    def load_model(self):
        """Load Torch2trt model."""
        from torch2trt import TRTModule
        model_trt = TRTModule()
        model_trt.load_state_dict(torch.load(self.model_path))
        model_trt.eval()
        return model_trt
    
    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the Torch2trt model."""
        with torch.no_grad():
            results = self.model(imgs)
            results = F.softmax(results, dim=1)  # probabilities
            results = results.cpu().numpy()
        return results

class JITClassifier(BaseClassifier):
    """
    JITClassifier.

    A class for loading and running classification models in jit format.
    """
    def __init__(self, model_path, device, size, class_names=None, fp16=False, verbose=False):
        """
        Initialize the JITClassifier class.

        Args:
            model_path (str): Path to the jit model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            verbose (bool): Whether to output the timer.
        """
        super().__init__(model_path, device, size, class_names, fp16, verbose)

    def load_model(self):
        """Load Torch2trt model."""
        model = torch.jit.load(self.model_path, map_location = self.device)
        model = model.half().eval() if self.fp16 else model.float().eval()
        return model

    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the jit model."""
        with torch.no_grad():
            results = self.model(imgs)
            results = F.softmax(results, dim=1)  # probabilities
            results = results.cpu().numpy()
        return results

# =============================================================================
# ClassifierBuilder
# =============================================================================   

@dataclass(frozen=True)
class ClassifierOptions:
    weights: str = './weights/yolov5s-cls.pt'
    size: Union[int, tuple] = 640
    class_names: dict = None
    fp16: bool = False
    verbose: bool = False

def Classifier(config: Union[str, ClassifierOptions]):
    """
    Initialize the Classifier function.

    Args:
        config (Union[str, ClassifierOptions]): A YAML configuration file path or a ClassifierOptions object.

    Raises:
        ValueError: If the config isn't a YAML file or ClassifierOptions object.
        ValueError: If the model weight file format is unsupported.

    Example:
        1. YAML file
            >>> model = Classifier("config.yaml")
            >>> results = model(images)
            >>> print(results)
        2. object
            >>> options = ClassifierOptions(conf_threshold=0.6, verbose=True....)
            >>> model = Classifier(options)
            >>> results = model(images)
            >>> print(results)
    """
    # Load config from a YAML file or a ClassifierOptions object
    if isinstance(config, str) and config.endswith(('yaml', 'yml')):
        with open(config, 'r') as f:
            setting = yaml.safe_load(f)

    elif isinstance(config, ClassifierOptions):
        setting = {k: v for k, v in config.__dict__.items() if not k.startswith("__")}

    else:
        raise ValueError("Config must be a YAML file or an object.")

    weight_path = setting['weights']
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    # Load the appropriate classifier model based on file type
    if weight_path.endswith('.onnx'):
        # print(f"Loading ONNX model: {weight_path}")
        model = ONNXClassifier(
            model_path=weight_path,
            device=device.type,  # 'cuda' or 'cpu'
            size=setting.get('size', 640),
            class_names=setting.get('class_names', None),
            fp16=setting.get('fp16', False),
            verbose=setting.get('verbose',False)
        )
    elif weight_path.endswith(('.pt', '.pth')):
        if '_trt' in weight_path:
            # print(f"Loading TensorRT model: {weight_path}")
            model = Torch2trtClassifier(
                model_path=weight_path,
                device=device.type,  # 'cuda' or 'cpu'
                size=setting.get('size', 640),
                class_names=setting.get('class_names', None),
                fp16=setting.get('fp16', False),
                verbose=setting.get('verbose',False)
            )
        else:
            # print(f"Loading Pytorch model: {weight_path}")
            model = TorchClassifier(
                model_path=weight_path,
                device=device.type,  # 'cuda' or 'cpu'
                size=setting.get('size', 640),
                class_names=setting.get('class_names', None),
                fp16=setting.get('fp16', False),
                verbose=setting.get('verbose',False)
            )
    elif weight_path.endswith('.jit'):
        # print(f"Loading JIT model: {weight_path}")
        model = JITClassifier(
            model_path=weight_path,
            device=device.type,  # 'cuda' or 'cpu'
            size=setting.get('size', 640),
            class_names=setting.get('class_names', None),
            fp16=setting.get('fp16', False),
            verbose=setting.get('verbose',False)
        )
    elif weight_path.endswith('.engine'):
        # print(f"Loading TensorRT model: {weight_path}")
        model = TensorRTClassifier(
            model_path=weight_path,
            device=device.type,  # 'cuda' or 'cpu'
            size=setting.get('size', 640),
            class_names=setting.get('class_names', None),
            fp16=setting.get('fp16', False),
            verbose=setting.get('verbose',False)
        )
    else:
        raise ValueError(f"Unsupported model format: {weight_path}")

    return model



import os
import sys
import cv2
import time
import yaml
import torch
import warnings
import torchvision
import numpy as np
from threading import Lock
from nxva.utilities import timer
from typing import Union, List, Optional
from dataclasses import dataclass


def DIoU(bboxes1, bboxes2):
    """
    Parameters
    ----------
    bboxes1 : numpy.ndarray (n, 4)
    bboxes2 : numpy.ndarray (m, 4)

    Returns
    -------
    dious : numpy.ndarray (n, m)
    """
    if bboxes1.ndim == 1:
        bboxes1 = np.expand_dims(bboxes1, 0)
    if bboxes2.ndim == 1:
        bboxes2 = np.expand_dims(bboxes2, 0)
    rows = bboxes1.shape[0]
    cols = bboxes2.shape[0]
    dious = np.zeros((rows, cols))
    if rows * cols == 0:
        return dious
    
    # xmin,ymin,xmax,ymax->[:,0],[:,1],[:,2],[:,3]
    w1 = bboxes1[:, 2:2+1] - bboxes1[:, 0:0+1]  # r
    h1 = bboxes1[:, 3:3+1] - bboxes1[:, 1:1+1]  # r
    w2 = bboxes2[:, 2:2+1] - bboxes2[:, 0:0+1]  # c
    h2 = bboxes2[:, 3:3+1] - bboxes2[:, 1:1+1]  # c
    
    area1 = w1 * h1  # r
    area2 = w2 * h2  # c

    center_x1 = (bboxes1[:, 2:2+1] + bboxes1[:, 0:0+1]) / 2  # r
    center_y1 = (bboxes1[:, 3:3+1] + bboxes1[:, 1:1+1]) / 2  # r
    center_x2 = (bboxes2[:, 2:2+1] + bboxes2[:, 0:0+1]) / 2  # c
    center_y2 = (bboxes2[:, 3:3+1] + bboxes2[:, 1:1+1]) / 2  # c
    # (r, c)
    center_dx = center_x1 - center_x2.T
    center_dy = center_y1 - center_y2.T
    
    # (r, c)
    inter_min_x = np.maximum(bboxes1[:, 0:0+1], bboxes2[:, 0:0+1].T)  
    inter_min_y = np.maximum(bboxes1[:, 1:1+1], bboxes2[:, 1:1+1].T) 
    inter_max_x = np.minimum(bboxes1[:, 2:2+1], bboxes2[:, 2:2+1].T)
    inter_max_y = np.minimum(bboxes1[:, 3:3+1], bboxes2[:, 3:3+1].T)
    outer_min_x = np.minimum(bboxes1[:, 0:0+1], bboxes2[:, 0:0+1].T)
    outer_min_y = np.minimum(bboxes1[:, 1:1+1], bboxes2[:, 1:1+1].T)
    outer_max_x = np.maximum(bboxes1[:, 2:2+1], bboxes2[:, 2:2+1].T)
    outer_max_y = np.maximum(bboxes1[:, 3:3+1], bboxes2[:, 3:3+1].T)
    
    # (r, c)
    inter_x = np.clip((inter_max_x - inter_min_x), a_min=0, a_max=None)
    inter_y = np.clip((inter_max_y - inter_min_y), a_min=0, a_max=None)
    inter_area = inter_x * inter_y
    outer_x = np.clip((outer_max_x - outer_min_x), a_min=0, a_max=None)
    outer_y = np.clip((outer_max_y - outer_min_y), a_min=0, a_max=None)
    inter_diag = center_dx ** 2 + center_dy ** 2
    outer_diag = outer_x ** 2 + outer_y ** 2
    union_area = area1 + area2.T - inter_area
    dious = inter_area / union_area - inter_diag / outer_diag
    dious = np.clip(dious, a_min=-1.0, a_max=1.0)
    
    return dious


# =============================================================================
# Non-maximum-suppression
# =============================================================================
def xywh2xyxy(x):
    # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
    return y


def box_area(box):
    # box = xyxy(4,n)
    return (box[2] - box[0]) * (box[3] - box[1])


def box_iou(box1, box2, eps=1e-7):
    # https://github.com/pytorch/vision/blob/master/torchvision/ops/boxes.py
    """
    Return intersection-over-union (Jaccard index) of boxes.
    Both sets of boxes are expected to be in (x1, y1, x2, y2) format.
    Arguments:
        box1 (Tensor[N, 4])
        box2 (Tensor[M, 4])
    Returns:
        iou (Tensor[N, M]): the NxM matrix containing the pairwise
            IoU values for every element in boxes1 and boxes2
    """

    # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
    (a1, a2), (b1, b2) = box1[:, None].chunk(2, 2), box2.chunk(2, 1)
    inter = (torch.min(a2, b2) - torch.max(a1, b1)).clamp(0).prod(2)

    # IoU = inter / (area1 + area2 - inter)
    return inter / (box_area(box1.T)[:, None] + box_area(box2.T) - inter + eps)


# =============================================================================
# Yolo detection tools
# =============================================================================
def letterbox(img, new_shape=(640, 640), color=(114, 114, 114), auto=True, scaleFill=False, scaleup=True):
    # Resize image to a 32-pixel-multiple rectangle https://github.com/ultralytics/yolov3/issues/232
    shape = img.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better test mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, 64), np.mod(dh, 64)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[1], new_shape[0])
        ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2
    
    if shape[::-1] != new_unpad:  # resize
        img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    
    return img, ratio, (dw, dh)


def clip_coords(boxes, img_shape):
    # Clip bounding xyxy bounding boxes to image shape (height, width)
    boxes[:, 0].clamp_(0, img_shape[1] - 1)  # x1
    boxes[:, 1].clamp_(0, img_shape[0] - 1)  # y1
    boxes[:, 2].clamp_(0, img_shape[1] - 1)  # x2
    boxes[:, 3].clamp_(0, img_shape[0] - 1)  # y2
    
    
def scale_coords(img1_shape, coords, img0_shape, ratio_pad=None):
    # Rescale coords (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    coords[:, [0, 2]] -= pad[0]  # x padding
    coords[:, [1, 3]] -= pad[1]  # y padding
    coords[:, :4] /= gain
    clip_coords(coords, img0_shape)
    return coords

class BaseDetector:
    """
    BaseDetector.

    A base class for creating detector.

    Attributes:
        model_path (str): Path to the PyTorch model file.
        device (str): "cuda" or "cpu".
        size (int): The size of the input image.
        conf_thres (float): Confidence threshold.
        iou_thres (float): IoU threshold.
        classes (list): List of class names.
        class_names (dict): Dictionary mapping class indices to class names.
        fp16 (bool): Whether to use FP16 precision.
        agnostic (bool): Merges overlapping boxes of different classes.
        verbose (bool): Whether to output the timer.
    """
    def __init__(self, model_path, device, size, conf_thres=0.25, iou_thres=0.7, classes=None, class_names=None, fp16=False, agnostic=False, verbose=False):
        """
        Initialize the BaseDetector class.

        Args:
            model_path (str): Path to the PyTorch model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            conf_thres (float): Confidence threshold.
            iou_thres (float): IoU threshold.
            classes (list): List of class names.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            agnostic (bool): Merges overlapping boxes of different classes.
            verbose (bool): Whether to output the timer.
        """
        self.model_path = model_path
        self.device = device
        self.size = size
        self.conf_thres = conf_thres
        self.iou_thres = iou_thres
        self.classes = classes
        self.class_names = class_names
        self.fp16 = fp16
        self.agnostic = agnostic
        self.verbose = verbose
        self.model = self.load_model()

        if self.class_names is None:
            if hasattr(self.model, 'module') and self.model.class_names:
                self.class_names = self.model.module.names
            elif hasattr(self.model, 'names') and self.model.names:
                self.class_names = self.model.names
            else:
                warnings.warn("No 'class_names' found in the configuration.", UserWarning)
        
    def load_model(self):
        raise NotImplementedError("Subclasses should overwrite this method to load the model!")
    
    @timer()
    def preprocess(self, imgs, to_tensor=False):
        """
        Prepares input image before inference.
        Args:
            imgs (list or np.ndarray): Input images.
            roi_list (list): A list of regions of ROI.
            to_tensor (bool): Converts the processed images into tensor format.
        """
        if not len(imgs):
            return [], [], [], []
        
        if isinstance(imgs, np.ndarray):
            imgs = [imgs]
        
        pre_imgs = []
        for img in imgs:
            pre_img = letterbox(img, self.size, auto=False)[0]
            pre_img = pre_img.transpose((2, 0, 1))[::-1]  # HWC -> CHW, BGR -> RGB
            pre_img = np.ascontiguousarray(pre_img)
            if to_tensor:
                pre_img = torch.from_numpy(pre_img).float().to(self.device)
                pre_img /= 255
                if self.fp16:
                    pre_img = pre_img.half()

            else:
                pre_img = pre_img.astype(np.float32) / 255.0
                if self.fp16:
                    pre_img = pre_img.astype(np.float16)

            pre_imgs.append(pre_img)

        if to_tensor:
            pre_imgs = torch.stack(pre_imgs)
        else:
            pre_imgs = np.stack(pre_imgs)

        return pre_imgs, imgs

    @timer()
    def post_process(self, pred, pre_imgs, imgs):
        """Post-processes predictions for an image and returns them."""
        dets = []
        for i, det in enumerate(pred):
            if len(det):
                det[:, :4] = scale_coords(pre_imgs.shape[2:], det[:, :4], imgs[i].shape).round()
                det = det.cpu().numpy()
                det = det[np.argsort(det[:, 0])]  # sort by horizontal axis
            else:
                det = det.cpu().numpy()

            dets.append(det)

        return dets
    
    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the specified model."""
        raise NotImplementedError("Subclasses should overwrite this method to run the detection!")
    
    def __call__(self, imgs):
        """Performs inference on the given image source."""
        pre_imgs, imgs = self.preprocess(imgs, to_tensor=self.needs_tensor())
        pred = self.infer(pre_imgs)
        if not self.needs_tensor():
            pred = torch.from_numpy(pred)

        pred = self.non_max_suppression(
            pred,
            conf_thres=self.conf_thres,
            iou_thres=self.iou_thres,
            classes=self.classes,
            agnostic=self.agnostic,
            max_det=300
        )

        dets = self.post_process(pred, pre_imgs, imgs)
        return dets
        
    def needs_tensor(self):
        return True
    
    def destroy(self):
        pass

    @timer()
    def non_max_suppression(
        self,
        prediction,
        conf_thres=0.25,
        iou_thres=0.7,
        classes=None,
        agnostic=False,
        multi_label=False,
        labels=(),
        max_det=300,
        nm=0,  # number of masks
    ):
        """Non-Maximum Suppression (NMS) on inference results to reject overlapping detections

        Returns:
            list of detections, on (n,6) tensor per image [xyxy, conf, cls]
        """

        if isinstance(prediction, (list, tuple)):  # YOLOv5 model in validation model, output = (inference_out, loss_out)
            prediction = prediction[0]  # select only inference output

        device = prediction.device
        mps = 'mps' in device.type  # Apple MPS
        if mps:  # MPS not fully supported yet, convert tensors to CPU before NMS
            prediction = prediction.cpu()
        bs = prediction.shape[0]  # batch size
        nc = prediction.shape[2] - nm - 5  # number of classes
        xc = prediction[..., 4] > conf_thres  # candidates

        # Checks
        assert 0 <= conf_thres <= 1, f'Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0'
        assert 0 <= iou_thres <= 1, f'Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0'

        # Settings
        # min_wh = 2  # (pixels) minimum box width and height
        max_wh = 7680  # (pixels) maximum box width and height
        max_nms = 30000  # maximum number of boxes into torchvision.ops.nms()
        time_limit = 0.5 + 0.05 * bs  # seconds to quit after
        redundant = True  # require redundant detections
        multi_label &= nc > 1  # multiple labels per box (adds 0.5ms/img)
        merge = False  # use merge-NMS

        t = time.time()
        mi = 5 + nc  # mask start index
        output = [torch.zeros((0, 6 + nm), device=prediction.device)] * bs
        for xi, x in enumerate(prediction):  # image index, image inference
            # Apply constraints
            # x[((x[..., 2:4] < min_wh) | (x[..., 2:4] > max_wh)).any(1), 4] = 0  # width-height
            x = x[xc[xi]]  # confidence

            # Cat apriori labels if autolabelling
            if labels and len(labels[xi]):
                lb = labels[xi]
                v = torch.zeros((len(lb), nc + nm + 5), device=x.device)
                v[:, :4] = lb[:, 1:5]  # box
                v[:, 4] = 1.0  # conf
                v[range(len(lb)), lb[:, 0].long() + 5] = 1.0  # cls
                x = torch.cat((x, v), 0)

            # If none remain process next image
            if not x.shape[0]:
                continue

            # Compute conf
            x[:, 5:] *= x[:, 4:5]  # conf = obj_conf * cls_conf

            # Box/Mask
            box = xywh2xyxy(x[:, :4])  # center_x, center_y, width, height) to (x1, y1, x2, y2)
            mask = x[:, mi:]  # zero columns if no masks

            # Detections matrix nx6 (xyxy, conf, cls)
            if multi_label:
                i, j = (x[:, 5:mi] > conf_thres).nonzero(as_tuple=False).T
                x = torch.cat((box[i], x[i, 5 + j, None], j[:, None].float(), mask[i]), 1)
            else:  # best class only
                conf, j = x[:, 5:mi].max(1, keepdim=True)
                x = torch.cat((box, conf, j.float(), mask), 1)[conf.view(-1) > conf_thres]

            # Filter by class
            if classes is not None:
                x = x[(x[:, 5:6] == torch.tensor(classes, device=x.device)).any(1)]

            # Apply finite constraint
            # if not torch.isfinite(x).all():
            #     x = x[torch.isfinite(x).all(1)]

            # Check shape
            n = x.shape[0]  # number of boxes
            if not n:  # no boxes
                continue
            elif n > max_nms:  # excess boxes
                x = x[x[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence
            else:
                x = x[x[:, 4].argsort(descending=True)]  # sort by confidence

            # Batched NMS
            c = x[:, 5:6] * (0 if agnostic else max_wh)  # classes
            boxes, scores = x[:, :4] + c, x[:, 4]  # boxes (offset by class), scores
            i = torchvision.ops.nms(boxes, scores, iou_thres)  # NMS
            if i.shape[0] > max_det:  # limit detections
                i = i[:max_det]
            if merge and (1 < n < 3E3):  # Merge NMS (boxes merged using weighted mean)
                # update boxes as boxes(i,4) = weights(i,n) * boxes(n,4)
                iou = box_iou(boxes[i], boxes) > iou_thres  # iou matrix
                weights = iou * scores[None]  # box weights
                x[i, :4] = torch.mm(weights, x[:, :4]).float() / weights.sum(1, keepdim=True)  # merged boxes
                if redundant:
                    i = i[iou.sum(1) > 1]  # require redundancy

            output[xi] = x[i]
            if mps:
                output[xi] = output[xi].to(device)
            if (time.time() - t) > time_limit:
                # LOGGER.warning(f'WARNING ⚠️ NMS time limit {time_limit:.3f}s exceeded')
                break  # time limit exceeded

        return output

class TorchDetector(BaseDetector):
    """
    TorchDetector.

    A class for loading and running detection models in PyTorch format.

    Attributes:
        nc (int): The number of classes in the model.
    """
    def __init__(self, model_path, device, size, conf_thres=0.25, iou_thres=0.7, classes=None, class_names=None, fp16=False, agnostic=False, verbose=False):
        """
        Initialize the TorchDetector class.

        Args:
            model_path (str): Path to the PyTorch model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            conf_thres (float): Confidence threshold.
            iou_thres (float): IoU threshold.
            classes (list): List of class names.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            agnostic (bool): Merges overlapping boxes of different classes.
            verbose (bool): Whether to output the timer.
        """
        super().__init__(model_path, device, size, conf_thres, iou_thres, classes, class_names, fp16, agnostic, verbose)
        self.nc = self.model.nc
        
    def load_model(self):
        """Load PyTorch model"""
        current_file_path = os.path.abspath(__file__)
        current_dir = os.path.dirname(current_file_path)
        sys.path.insert(0, os.path.join(current_dir, 'yolov5'))
        from models.experimental import attempt_load
        sys.path.remove(os.path.join(current_dir, 'yolov5'))

        model = attempt_load(self.model_path, device=self.device)
        model = model.half().eval() if self.fp16 else model.float().eval()
        return model
    
    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the PyTorch model."""
        with torch.no_grad():
            pred = self.model(imgs)[0]
        return pred
    

class ONNXDetector(BaseDetector):
    """
    ONNXDetector.

    A class for loading and running detection models in ONNX format.

    Attributes:
        size (int): The size of the model.
        nc (int): The number of classes in the model.
    """
    def __init__(self, model_path, device, size, conf_thres=0.25, iou_thres=0.7, classes=None, class_names=None, fp16=False, agnostic=False, verbose=False):
        """
        Initialize the ONNXDetector class.

        Args:
            model_path (str): Path to the ONNX model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            conf_thres (float): Confidence threshold.
            iou_thres (float): IoU threshold.
            classes (list): List of class names.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            agnostic (bool): Merges overlapping boxes of different classes.
            verbose (bool): Whether to output the timer.
        """
        super().__init__(model_path, device, size, conf_thres, iou_thres, classes, class_names, fp16, agnostic, verbose)
        self.size = self.determine_input_size()
        assert self.size == (size, size), "The input size of the model is not the same as the size in the config file."
        self.nc = self._get_num_classes()
        
    def load_model(self):
        """Load ONNX model"""
        import onnxruntime as ort
        
        if self.device == "cuda":
            print("Using CUDA")
            providers = [("CUDAExecutionProvider", {"device_id": 0})]  
        else:
            print("Using CPU")
            providers = ["CPUExecutionProvider"]
            
        return ort.InferenceSession(self.model_path, providers=providers)
    
    def determine_input_size(self):
        """Determines the input size expected by the model."""
        input_shape = self.model.get_inputs()[0].shape
        return (input_shape[2], input_shape[3])
    
    def _get_num_classes(self):
        """Determines the number of classes by the model."""
        outputs = self.model.get_outputs()
        for output in outputs:
            return output.shape[2] - 5
    
    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the ONNX model."""
        pred = self.model.run(None, {self.model.get_inputs()[0].name: imgs})[0]
        return pred
    
    def needs_tensor(self):
        return False

class TensorRTDetector_Threading(BaseDetector):
    """
    This class is used for TensorRT inference in sub-threading. It uses a dedicated CUDA context for each thread.
    context = engine.create_execution_context() default creates a context in main threading.
    We need to actively create a context manager in each sub-thread by using cuda.Device(0).make_context().
    Seems like to register a context in each thread, we need to use a lock to avoid conflicts.
    """
    def __init__(self, model_path, device, size, conf_thres=0.25, iou_thres=0.7, classes=None, class_names=None, fp16=False, agnostic=False, verbose=False):
        assert device == "cuda", "TensorRT only supports CUDA."
        self.trt, self.cuda = self._import_dependencies()
        super().__init__(model_path, device, size, conf_thres, iou_thres, classes, class_names, fp16, agnostic, verbose)

        # create a dedicated cuda context
        self.cfx = self.cuda.Device(0).make_context()
        
        self.engine = self.model
        self.context = self.engine.create_execution_context()
        self.size, self.fixed_batch = self.get_input_size()
        assert self.size == (size, size), "The input size of the model is not the same as the size in the config file."
        self.bindings, self.inputs, self.outputs, self.stream = self.prepare_buffers()
        self.lock = Lock()
        self.nc = self._get_num_classes()
        
    def _import_dependencies(self):
        import tensorrt as trt
        import pycuda.driver as cuda
        import pycuda.autoinit  # Ensures initialization of CUDA
        return trt, cuda
    
    def get_input_size(self):
        """Determines the input size expected by the model."""
        for binding_index in range(self.engine.num_bindings):
            if self.engine.binding_is_input(binding_index):
                shape = self.engine.get_binding_shape(binding_index)
                return (shape[2], shape[3]), shape[0]
    
    def load_model(self):
        logger = self.trt.Logger(self.trt.Logger.INFO)
        with open(self.model_path, "rb") as f, self.trt.Runtime(logger) as runtime:
            return runtime.deserialize_cuda_engine(f.read())
        
    def prepare_buffers(self):
        """Prepare input and output buffers for TensorRT inference."""
        bindings = [] 
        stream = self.cuda.Stream() 
        inputs, outputs = [], [] 

        for binding_index in range(self.engine.num_bindings): 
            binding_shape = self.engine.get_binding_shape(binding_index) 
            if self.fp16 and self.engine.get_binding_dtype(binding_index) == self.trt.DataType.FLOAT: 
                binding_dtype = np.float16 
            else:
                binding_dtype = self.trt.nptype(self.engine.get_binding_dtype(binding_index)) 
            binding_size = self.trt.volume(binding_shape) * np.dtype(binding_dtype).itemsize 
            device_mem = self.cuda.mem_alloc(binding_size) 
            bindings.append(int(device_mem)) 
            
            if self.engine.binding_is_input(binding_index): 
                host_mem = np.zeros(binding_shape, dtype=binding_dtype)
                inputs.append({'host': host_mem, 'device': device_mem})
            else:
                host_mem = np.zeros(binding_shape, dtype=binding_dtype)
                outputs.append({'host': host_mem, 'device': device_mem})

        return bindings, inputs, outputs, stream
    
    def _get_num_classes(self):
        """Determines the number of classes by the model."""
        for binding_index in range(self.engine.num_bindings):
            if not self.engine.binding_is_input(binding_index):
                binding_shape = self.engine.get_binding_shape(binding_index)
                return binding_shape[2] - 5
    
    @timer()
    def infer(self, imgs):
        with self.lock:
            try:
                self.cfx.push()  # attach the context before running the engine
                batch_size = imgs.shape[0]
                if batch_size > self.fixed_batch:
                    raise ValueError(f"Input batch size {batch_size} exceeds the fixed batch size {self.fixed_batch}.")
                if batch_size < self.fixed_batch:
                    padding = np.zeros((self.fixed_batch - batch_size, *imgs.shape[1:]), dtype=imgs.dtype)
                    imgs = np.concatenate([imgs, padding], axis=0) 

                # Transfer input data to the GPU
                self.cuda.memcpy_htod_async(self.inputs[0]['device'], imgs, self.stream)

                # Run inference
                self.context.execute_async_v2(bindings=self.bindings, stream_handle=self.stream.handle)

                # Transfer predictions back from the GPU
                for output in self.outputs:
                    self.cuda.memcpy_dtoh_async(output['host'], output['device'], self.stream)

                self.stream.synchronize()
                pred = self.outputs[0]['host'][0:batch_size]
            except self.cuda.LogicError as e:
                print(f"CUDA logic error: {str(e)}")
            except self.cuda.MemoryError as e:
                print(f"CUDA memory error: {str(e)}")
            except Exception as e:
                print(f"General error during CUDA operations: {str(e)}")
            finally:  # detach the context after running the engine
                self.cfx.pop()
                
        return pred
    
    def destroy(self):
        self.cfx.pop() # detach the context before destroying the engine
    
    def needs_tensor(self):
        return False

class TensorRTDetector(BaseDetector):
    """
    TensorRTDetector.

    A class for loading and running detection models in TensorRT format.

    Attributes:
        trt (module): TensorRT module.
        cuda (module): PyCUDA module.
        engine (TensorRT.ICudaEngine): TensorRT engine.
        context (TensorRT.IExecutionContext): TensorRT execution context.
        size (tuple): The size of the input image.
        fixed_batch (int): The fixed batch size of the model.
        bindings (list): List of buffer bindings. Pointer to the input and output buffers.
        inputs (list): List of input buffers.  
        outputs (list): List of output buffers.
        stream (pycuda.driver.Stream): CUDA stream. Used for asynchronous memory transfers.
        nc (int): The number of classes in the model
    """
    def __init__(self, model_path, device, size, conf_thres=0.25, iou_thres=0.7, classes=None, class_names=None, fp16=False, agnostic=False, verbose=False):
        """
        Args:
            model_path (str): Path to the TensorRT model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            conf_thres (float): Confidence threshold.
            iou_thres (float): IoU threshold.
            classes (list): List of class names.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            agnostic (bool): Merges overlapping boxes of different classes.
            verbose (bool): Whether to output the timer.
        """
        assert device == "cuda", "TensorRT only supports CUDA."
        self.trt, self.cuda = self._import_dependencies()
        super().__init__(model_path, device, size, conf_thres, iou_thres, classes, class_names, fp16, agnostic, verbose)
        self.engine = self.model
        self.context = self.engine.create_execution_context()  # create a context
        self.size, self.fixed_batch = self.get_input_size()
        assert self.size == (size, size), "The input size of the model is not the same as the size in the config file."
        self.bindings, self.inputs, self.outputs, self.stream = self.prepare_buffers()
        self.nc = self._get_num_classes()

    def _import_dependencies(self):
        import tensorrt as trt
        import pycuda.driver as cuda
        import pycuda.autoinit  # Ensures initialization of CUDA
        return trt, cuda
    
    def get_input_size(self):
        """Determines the input size expected by the model."""
        for binding_index in range(self.engine.num_bindings):
            if self.engine.binding_is_input(binding_index):
                shape = self.engine.get_binding_shape(binding_index)
                return (shape[2], shape[3]), shape[0]
    
    def load_model(self):
        """Load TensorRT model"""
        logger = self.trt.Logger(self.trt.Logger.INFO)
        with open(self.model_path, "rb") as f, self.trt.Runtime(logger) as runtime:
            return runtime.deserialize_cuda_engine(f.read())
        
    def prepare_buffers(self):
        """Prepare input and output buffers for TensorRT inference."""
        bindings = [] 
        stream = self.cuda.Stream() 
        inputs, outputs = [], [] 

        for binding_index in range(self.engine.num_bindings): 
            binding_shape = self.engine.get_binding_shape(binding_index) 
            if self.fp16 and self.engine.get_binding_dtype(binding_index) == self.trt.DataType.FLOAT: 
                binding_dtype = np.float16 
            else:
                binding_dtype = self.trt.nptype(self.engine.get_binding_dtype(binding_index)) 
            binding_size = self.trt.volume(binding_shape) * np.dtype(binding_dtype).itemsize 
            device_mem = self.cuda.mem_alloc(binding_size) 
            bindings.append(int(device_mem)) 
            
            if self.engine.binding_is_input(binding_index): 
                host_mem = np.zeros(binding_shape, dtype=binding_dtype)
                inputs.append({'host': host_mem, 'device': device_mem})
            else:
                host_mem = np.zeros(binding_shape, dtype=binding_dtype)
                outputs.append({'host': host_mem, 'device': device_mem})

        return bindings, inputs, outputs, stream
    
    def _get_num_classes(self):
        """Determines the number of classes by the model."""
        for binding_index in range(self.engine.num_bindings):
            if not self.engine.binding_is_input(binding_index):
                binding_shape = self.engine.get_binding_shape(binding_index)
                nc = binding_shape[2] - 5
                return nc
    
    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the TensorRT model."""
        try:
            batch_size = imgs.shape[0]
            if batch_size > self.fixed_batch:
                raise ValueError(f"Input batch size {batch_size} exceeds the fixed batch size {self.fixed_batch}.")
            if batch_size < self.fixed_batch:
                padding = np.zeros((self.fixed_batch - batch_size, *imgs.shape[1:]), dtype=imgs.dtype)
                imgs = np.concatenate([imgs, padding], axis=0) 

            # Transfer input data to the GPU
            self.cuda.memcpy_htod_async(self.inputs[0]['device'], imgs, self.stream)

            # Run inference
            self.context.execute_async_v2(bindings=self.bindings, stream_handle=self.stream.handle)

            # Transfer predictions back from the GPU
            for output in self.outputs:
                self.cuda.memcpy_dtoh_async(output['host'], output['device'], self.stream)

            self.stream.synchronize()
            pred = self.outputs[0]['host'][0:batch_size]
        except self.cuda.LogicError as e:
            print(f"CUDA logic error: {str(e)}")
        except self.cuda.MemoryError as e:
            print(f"CUDA memory error: {str(e)}")
        except Exception as e:
            print(f"General error during CUDA operations: {str(e)}")

        return pred
    
    def needs_tensor(self):
        return False

class Torch2trtDetector(BaseDetector):
    """
    Torch2trtDetector.

    A class for loading and running detection models in trt format.

    Attributes:
        nc (int): The number of classes in the model.
    """
    def __init__(self, model_path, device, size, conf_thres=0.25, iou_thres=0.7, classes=None, class_names=None, fp16=False, agnostic=False, verbose=False):
        """
        Initialize the Torch2trtDetector class.

        Args:
            model_path (str): Path to the Torch2trt model file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            conf_thres (float): Confidence threshold.
            iou_thres (float): IoU threshold.
            classes (list): List of class names.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            agnostic (bool): Merges overlapping boxes of different classes.
            verbose (bool): Whether to output the timer.
        """
        super().__init__(model_path, device, size, conf_thres, iou_thres, classes, class_names, fp16, agnostic, verbose)
        self.nc = self._get_num_classes()
        
    def load_model(self):
        """Load Torch2trt model"""
        from torch2trt import TRTModule
        model_trt = TRTModule()
        model_trt.load_state_dict(torch.load(self.model_path))
        model_trt.eval()
        return model_trt
    
    def _get_num_classes(self):
        """Determines the number of classes by the model."""
        engine = self.model.engine
        for binding_index in range(engine.num_bindings):
            if not engine.binding_is_input(binding_index):
                binding_shape = engine.get_binding_shape(binding_index)
                nc = binding_shape[2] - 5
                
        return nc
    
    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the Torch2trt model."""
        with torch.no_grad():
            pred = self.model(imgs)[0]
        return pred
    
class JITDetector(BaseDetector):
    """
    JITDetector.

    A class for loading and running detection models in jit format.

    Attributes:
        nc (int): The number of classes in the model.
    """
    def __init__(self, model_path, device, size, conf_thres=0.25, iou_thres=0.7, classes=None, class_names=None, fp16=False, agnostic=False, verbose=False):
        """
        Initialize the JITDetector class.

        Args:
            model_path (str): Path to the JIT file.
            device (str): "cuda" or "cpu".
            size (int): The size of the input image.
            conf_thres (float): Confidence threshold.
            iou_thres (float): IoU threshold.
            classes (list): List of class names.
            class_names (dict): Dictionary mapping class indices to class names.
            fp16 (bool): Whether to use FP16 precision.
            agnostic (bool): Merges overlapping boxes of different classes.
            verbose (bool): Whether to output the timer.
        """
        super().__init__(model_path, device, size, conf_thres, iou_thres, classes, class_names, fp16, agnostic, verbose)
        self.nc = self._get_num_classes()

    def load_model(self):
        """Load JIT model"""
        model = torch.jit.load(self.model_path, map_location = self.device)
        model = model.half().eval() if self.fp16 else model.float().eval()
        return model
    
    def _get_num_classes(self):
        """
        Determines the number of classes by the model.

        state_dict (OrderedDict):
            'model.0.conv.weight': torch.Size([32, 3, 6, 6])
            'model.0.conv.bias': torch.Size([32])
             .....
            'model.24.m.1.weight': torch.Size([255, 256, 1, 1])
            'model.24.m.1.bias': torch.Size([255])
            'model.24.m.2.weight': torch.Size([255, 512, 1, 1])
            'model.24.m.2.bias': torch.Size([255])
        """
        num_anchors = 3
        state_dict = self.model.state_dict()
        for name, param in state_dict.items():
            if 'm.0.weight' in name:
                output_channels  = param.shape[0]
                nc = (output_channels // num_anchors) - 5
                return nc
        
    @timer()
    def infer(self, imgs):
        """Runs inference on a given image using the JIT model."""
        with torch.no_grad():
            pred = self.model(imgs)[0]
            return pred
        
# =============================================================================
# DetectorBuilder
# =============================================================================

@dataclass(frozen=True)
class DetectorOptions:
    weights: str = './weights/yolov5s.pt'
    size: int = 640
    conf_threshold: float = 0.25
    iou_threshold: float = 0.7
    classes: Optional[List[int]] = None
    class_names: dict = None
    fp16: bool = False
    agnostic_nms: bool = False
    verbose: bool = False

def Detector(config: Union[str, DetectorOptions]):
    """
    Initialize the Detector function.

    Args:
        config (Union[str, DetectorOptions]): A YAML configuration file path or a DetectorOptions object.
    Raises:
        ValueError: If the config isn't a YAML file or 'DetectorOptions' object.
        ValueError: If the model weight file format is unsupported.

    Example:
        1. YAML file
            >>> model = Detector("config.yaml")
            >>> results = model(images)
            >>> print(results)
        2. object
            >>> options = DetectorOptions(conf_threshold=0.6, verbose=True....)
            >>> model = Detector(options)
            >>> results = model(images)
            >>> print(results)
    """
    # Load config from a YAML file or a DetectorOptions object
    if isinstance(config, str) and config.endswith(('yaml', 'yml')):
        with open(config, 'r') as f:
            setting = yaml.safe_load(f)

    elif isinstance(config, DetectorOptions):
        setting = {k: v for k, v in config.__dict__.items() if not k.startswith("__")}

    else:
        raise ValueError("Config must be a YAML file or an object.")

    weight_path = setting['weights']
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    # Load the appropriate detector model based on file type
    if weight_path.endswith('.onnx'):
        # print(f"Loading ONNX model: {weight_path}")
        model = ONNXDetector(
            model_path=weight_path,
            device=device.type,  # 'cuda' or 'cpu'
            size=setting.get('size', 640),
            conf_thres=setting.get('conf_threshold', 0.25),
            iou_thres=setting.get('iou_threshold', 0.7),
            classes=setting.get('classes', None),
            class_names = setting.get('class_names', None),
            fp16=setting.get('fp16', False),
            agnostic = setting.get('agnostic_nms', False),
            verbose=setting.get('verbose', False)
        )
    elif weight_path.endswith(('.pt', '.pth')):
        if '_trt' in weight_path:
            # print(f"Loading TensorRT model: {weight_path}")
            model = Torch2trtDetector(
                model_path=weight_path,
                device=device.type,  # 'cuda' or 'cpu'
                size=setting.get('size', 640),
                conf_thres=setting.get('conf_threshold', 0.25),
                iou_thres=setting.get('iou_threshold', 0.7),
                classes=setting.get('classes', None),
                class_names = setting.get('class_names', None),
                fp16=setting.get('fp16', False),
                agnostic = setting.get('agnostic_nms', False),
                verbose=setting.get('verbose', False)
            )
        else:
            # print(f"Loading Pytorch model: {weight_path}")
            model = TorchDetector(
                model_path=weight_path,
                device=device,
                size=setting.get('size', 640),
                conf_thres=setting.get('conf_threshold', 0.25),
                iou_thres=setting.get('iou_threshold', 0.7),
                classes=setting.get('classes', None),
                class_names = setting.get('class_names', None),
                fp16=setting.get('fp16', False),
                agnostic = setting.get('agnostic_nms', False),
                verbose=setting.get('verbose', False)
            )
    elif weight_path.endswith('.engine'):
        # print(f"Loading TensorRT model: {weight_path}")
        model = TensorRTDetector(
            model_path=weight_path,
            device=device.type,  # 'cuda' or 'cpu'
            size=setting.get('size', 640),
            conf_thres=setting.get('conf_threshold', 0.25),
            iou_thres=setting.get('iou_threshold', 0.7),
            classes=setting.get('classes', None),
            class_names = setting.get('class_names', None),
            fp16=setting.get('fp16', False),
            agnostic = setting.get('agnostic_nms', False),
            verbose=setting.get('verbose', False)
        )
    elif weight_path.endswith('.jit'):
        # print(f"Loading JIT model: {weight_path}")
        model = JITDetector(
            model_path=weight_path,
            device=device.type,  # 'cuda' or 'cpu'
            size=setting.get('size', 640),
            conf_thres=setting.get('conf_threshold', 0.25),
            iou_thres=setting.get('iou_threshold', 0.7),
            classes=setting.get('classes', None),
            class_names = setting.get('class_names', None),
            fp16=setting.get('fp16', False),
            agnostic = setting.get('agnostic_nms', False),
            verbose=setting.get('verbose', False)
        )
    
    else:
        raise ValueError(f"Unsupported model format: {weight_path}")
    
    return model

# =============================================================================
# Load model
# =============================================================================
    
def load_model(weight, device):
    current_file_path = os.path.abspath(__file__)
    current_dir = os.path.dirname(current_file_path)
    sys.path.insert(0, os.path.join(current_dir, 'yolov5'))
    from models.experimental import attempt_load
    sys.path.remove(os.path.join(current_dir, 'yolov5'))

    model = attempt_load(weight, device=device)
    return model



from .detection import Detector, DetectorOptions, load_model
from .classification import Classifier, ClassifierOptions
from .pose_estimation import PoseEstimator, PoseEstimatorOptions
from .modelconvert import ModelConverter, torch2onnx


import os
import gc
import cv2
cv2.setNumThreads(0)
import time
import yaml
import platform
import threading
import numpy as np       
        
def is_jetson():
    """Check if the code is running on an NVIDIA Jetson platform."""
    compat_file_path = "/proc/device-tree/compatible"
    if os.path.isfile(compat_file_path):
        with open(compat_file_path, "r") as file:
            compat_str = file.read()
            return "nvidia" in compat_str and "tegra" in compat_str
    return False


class MultiStreaming(object):
    """
    A class to handle multi-camera streaming with automatic reconnection, frame resizing and gstreamer support.
    """
    def __init__(self, 
                 config=None, 
                 reconnect_interval=30,
                 verbose=False):
        """
        Initialize the MultiStreaming object with the given configuration settings.
        
        Args:
            config (str or dict): Configuration settings for the cameras, 
                which can be provided as a path to a YAML file or directly as a dictionary. 
                The configuration should include the following sections:
                
                'parameter': Dictionary containing:
                    - 'WIDTH' (int): Width of the frame.
                    - 'HEIGHT' (int): Height of the frame.
                    - 'GSTREAMER' (bool): Whether to use GStreamer for streaming.
                    - 'CODEC' (str, optional): Codec to use, default is 'h264'.
                
                'cameras': List of dictionaries, each containing:
                    - 'ip' (str): IP address of the camera.
                    - 'user' (str, optional): Username for the camera.
                    - 'pw' (str, optional): Password for the camera.
                
                Example configuration:
                ```
                parameter:
                    WIDTH: 1920
                    HEIGHT: 1080
                    GSTREAMER: True
                    CODEC: h264
                cameras:
                    - ip: '192.168.1.100'
                      user: 'admin'
                      pw: 'password'
                ```

            reconnect_interval (int): Interval in seconds to attempt reconnection to a camera.
            verbose (bool): Enable verbose output to help with debugging.
        """
        self.reconnect_interval = reconnect_interval
        self.verbose = verbose
        
        if config:
            self._load_and_validate_config(config)
            self._generate_rtsp_strings()
        else:
            raise ValueError('Configuration must be provided, either as a dictionary or a YAML file path.')
        
        self.running = True
        self.mutex = threading.Lock()
        self.caps = [cv2.VideoCapture() for _ in range(self.num)]
        self.caps_info = [None for _ in range(self.num)]
        self.frames = [np.zeros((self.h, self.w, 3), dtype=np.uint8) for i in range(self.num)]
        self.status = [False] * self.num
        self.rtsp_reconnect = [None] * self.num

    def _load_and_validate_config(self, config):
        if isinstance(config, str):
            assert config.endswith(('.yaml', '.yml'))
            with open(config, 'r') as f:
                self.setting = yaml.safe_load(f)
        elif isinstance(config, dict):
            self.setting = config
        else:
            raise TypeError('Config must be provided as a yaml file or dictionary.')
        
        self.num = len(self.setting['cameras'])
        for i in range(self.num):
            rtsp = self.setting['cameras'][i].get('rtsp', None)
            if not rtsp:
                self.setting['cameras'][i]['rtsp'] = [
                    {'ip': self.setting['cameras'][i]['ip'],
                     'user': self.setting['cameras'][i].get('user', None),
                     'pw': self.setting['cameras'][i].get('pw', None)}
                ]
        
        # Ensure 'parameter' section exists and set default values if necessary
        param_defaults = {'WIDTH': 1920, 'HEIGHT': 1080, 'GSTREAMER': False, 'CODEC': 'h264'}
        self.setting.setdefault('parameter', {})
        for key, default in param_defaults.items():
            self.setting['parameter'].setdefault(key, default)

        # Ensure 'cameras' section exists and is a list
        if 'cameras' not in self.setting or not isinstance(self.setting['cameras'], list):
            raise ValueError("The 'cameras' section must be a list and cannot be empty.")

        # Extract values to simplify further usage
        self.w = self.setting['parameter']['WIDTH']
        self.h = self.setting['parameter']['HEIGHT']
        self.gst_enable = self.setting['parameter']['GSTREAMER']
        self.codec = self.setting['parameter']['CODEC']

    def _generate_rtsp_strings(self):
        # Gstreamer pipeline
        if self.gst_enable:
            cpu_arch = platform.machine()
            if cpu_arch == 'x86_64':
                gst_pipeline = [
                    f'queue max-size-buffers=0 max-size-bytes={1024*1024} leaky=downstream',  # 1MB buffer to keep GOP not be truncated
                    f'rtp{self.codec}depay', 
                    f'{self.codec}parse', 
                    f'avdec_{self.codec}', 
                    'videoconvert n-threads=1', 
                    'appsink drop=true sync=false'
                ]
                # gst_pipeline = [
                #     f'queue max-size-buffers=0 max-size-bytes={1024*1024} leaky=downstream',
                #     f'rtp{self.codec}depay', 
                #     f'{self.codec}parse', 
                #     f'nv{self.codec}dec',
                #     'nvvideoconvert'  # it is important to use nvvideoconvert after nv{codec}dec for performance
                #     'queue max-size-buffers=1 leaky=downstream',
                #     'videoconvert n-threads=1',  # it is important to add convert after nvvideoconvert for performance
                #     'video/x-raw, format=RGBA',
                #     'appsink caps="video/x-raw,format=BGR drop=true sync=false'
                # ]
            elif cpu_arch == 'aarch64' and is_jetson():
                gst_pipeline = [
                    f'queue max-size-buffers=0 max-size-bytes={1024*1024} leaky=downstream',
                    f'rtp{self.codec}depay', 
                    f'{self.codec}parse config-interval=1', 
                    'nvv4l2decoder',  #  drop-frame has some problem, disable-dpb=true need the stream with no B-frames
                    'nvvidconv', 
                    'queue max-size-buffers=1 leaky=downstream',
                    'video/x-raw,format=RGBA',          # explicitly set RGBA format for system memory
                    'videoconvert n-threads=1',         
                    'video/x-raw,format=BGR',           # cap filter, ensure the output is in BGR format
                    'appsink drop=true sync=false'
                ]
                # gst_pipeline = [
                #     f'rtp{self.codec}depay', f'{self.codec}parse', 'nvv4l2decoder', 
                #     'nvvidconv', 'appsink drop=true sync=false'
                # ]
            else:
                gst_pipeline = []
                print('Not allow to use Gstreamer under the current platform! (Only support x86_64 and Jetson)')
        
        # Collect RTSP strings
        self.rtsp_strings = [[] for _ in range(self.num)]

        for i in range(len(self.setting['cameras'])):
            camera = self.setting['cameras'][i]
            for rtsp in camera['rtsp']:
                ip = rtsp['ip']
                user = rtsp['user']
                pw = rtsp['pw']
                if self.gst_enable and gst_pipeline:
                    # Gstreamer pipeline
                    rtspsrc_base  = f'rtspsrc location=rtsp://{ip}'
                    if user and pw:
                        rtspsrc_base += f' user-id={user} user-pw={pw}'

                    # https://forums.developer.nvidia.com/t/deepstream-losing-relevance-because-of-outdated-dependecies/203426/7?utm_source=chatgpt.com 
                    rtspsrc_params = [
                        'short-header=true',
                        'is-live=true',
                        'protocols=tcp',
                        'do-rtsp-keep-alive=true',
                        'timeout=5000000',  # 5s
                        'tcp-timeout=5000000'
                        # 'latency=200',
                        # 'drop-on-latency=true',
                        # 'timeout=10',
                        # 'tcp-timeout=5',
                        # 'retry=3',
                        # 'do-retransmission=false'
                    ]
                    rtspsrc = rtspsrc_base + ' ' + ' '.join(rtspsrc_params)
                    
                    # Gstreamer pipeline put in front of the default rtsp
                    if len(self.rtsp_strings[i]) > 1:
                        index = len(self.rtsp_strings[i])//2
                        self.rtsp_strings[i].insert(index, ' ! '.join([rtspsrc] + gst_pipeline))
                    else:
                        self.rtsp_strings[i].append(' ! '.join([rtspsrc] + gst_pipeline))

                # if Gstreamer is not used, use the default rtsp
                if user and pw:
                    self.rtsp_strings[i].append(f'rtsp://{user}:{pw}@{ip}')
                else:
                    self.rtsp_strings[i].append(f'rtsp://{ip}')
        
    def __len__(self):
        return self.num
    
    def init_cameras(self, i):
        """Initialize all cameras based on the configuration settings."""
        for rtsp in self.rtsp_strings[i]:
            self.caps[i] = self._open_camera(rtsp, self.gst_enable)
            self.caps_info[i] = self._get_info(self.caps[i])
            
            if self.caps[i].isOpened():
                self.status[i] = True
                self.rtsp_reconnect[i] = rtsp
                if self.verbose:
                    print(f'camera {i}:', self.caps_info[i], flush=True)
                break
            else:
                self.status[i] = False
                if self.verbose:
                    print(f'camera {i}: failed to open initially', flush=True)

    def stop(self):
        """Stop all camera threads."""
        self.running = False

    def release(self):
        """Stop all camera threads and release resources."""
        self.running = False
        time.sleep(0.1)
        for cap in self.caps:
            if cap.isOpened():
                cap.release()
        # Wait for threads to finish
        for thread in self.threads:
            thread.join()
        
    def run(self):
        """Start the camera streaming threads."""
        self.threads = []
        for i in range(self.num):
            t = threading.Thread(
                target=self._run_one,
                args=(i, ),
                daemon=True
                )
            t.start()
            self.threads.append(t)
        
    def get_frames(self, status=False):
        """
        Retrieve the latest frames from all cameras. 
        Optionally include status.
        """
        with self.mutex:
            frames = self.frames.copy()
            if status:
                return frames, self.status.copy()
            return frames
            
    def _run_one(self, i):
        """Handle the streaming of one individual camera."""
        self.init_cameras(i)
        cap = self.caps[i]
        resize = (self.caps_info[i]['w'], self.caps_info[i]['h']) != (self.w, self.h)
        reconnect_time = time.time() - self.reconnect_interval
        
        try:
            fail = 0
            while self.running:
                # successfully read frame
                if cap.isOpened():
                    ret, frame = cap.read()
                    if ret:
                        if self.verbose and not self.status[i]:
                                print(f'camera {i} reconnects successfully', flush=True)
                        if resize:
                            frame = cv2.resize(frame, (self.w, self.h))
                        with self.mutex:
                            self.frames[i] = frame
                            self.status[i] = True
                        time.sleep(0.02)
                        continue
                    else:
                        fail += 1
                else:
                    fail += 1
                
                # camera is not open or failed to read frame
                with self.mutex:
                    self.frames[i] = np.zeros((self.h, self.w, 3), dtype=np.uint8)
                    self.status[i] = False
                    
                if time.time() - reconnect_time > self.reconnect_interval and fail > 10:
                    reconnect_time = time.time()
                    if self.verbose:
                        print(f'camera {i} is not open, try reconnecting', flush=True)
                    if cap.isOpened():
                        cap.release()
                    del cap
                    gc.collect()
                    time.sleep(0.2)  
                    if self.rtsp_reconnect[i]:
                        cap = self._open_camera(self.rtsp_reconnect[i], self.gst_enable)
                        if not cap.isOpened():
                            self.rtsp_reconnect[i] = None
                    else:
                        for rtsp in self.rtsp_strings[i]:
                            cap = self._open_camera(rtsp, self.gst_enable)
                            if cap.isOpened():
                                self.rtsp_reconnect[i] = rtsp
                                break
                    self.caps[i] = cap
                    self.caps_info[i] = self._get_info(self.caps[i])
                    resize = (self.caps_info[i]['w'], self.caps_info[i]['h']) != (self.w, self.h)
                    fail = 0
                
                time.sleep(0.02)
        finally:
            if cap.isOpened():
                cap.release()
    
    @staticmethod
    def _open_camera(rtsp, gst):
        """Open a camera using the given RTSP stream and settings for GStreamer."""
        def _open(cap):
            cap.open(rtsp, cv2.CAP_GSTREAMER) if gst else cap.open(rtsp)
                
        cap = cv2.VideoCapture()
        t = threading.Thread(
            target=_open, 
            args=(cap,), 
            daemon=True
            )
        t.start()
        t.join(15)
        
        return cap
    
    @staticmethod
    def _get_info(cap):
        """Return camera properties such as width, height, FPS, and codec."""
        try:
            if cap.isOpened():
                w = cap.get(cv2.CAP_PROP_FRAME_WIDTH)
                h = cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
                fps = cap.get(cv2.CAP_PROP_FPS)
                fourcc = int(cap.get(cv2.CAP_PROP_FOURCC))
                codec_format = ''.join([chr((fourcc >> 8 * i) & 0xFF) for i in range(4)])
                return {
                    'w': int(w),
                    'h': int(h),
                    'fps': int(fps),
                    'codec': codec_format
                }
        except Exception as e:
            print(f"Error getting camera properties: {e}", flush=True)

        return {
            'w': 0,
            'h': 0,
            'fps': 0,
            'codec': ''
        }


import numpy as np
import time

from .track import KalmanBoxTracker
from .reid_multibackend import ReIDDetectMultiBackend
from .utils.tools import nn_cosine_distance, xyxy2xywh, k_previous_obs, get_speed_direction
from .utils.association import iou_batch, giou_batch, ciou_batch, diou_batch, ct_dist, associate, linear_assignment
import torch

ASSO_FUNCS = {  "iou": iou_batch, # IoU計算方式選擇(in association.py)
    "giou": giou_batch,
    "ciou": ciou_batch,
    "diou": diou_batch,
    "ct_dist": ct_dist}

class NexuniSort(): #參照OCSort的架構, 融合StrongSort的特徵計算(不用特徵的話=只用IoU重疊來匹配)
    def __init__(self,
            settings):
        self.trackers = [] #儲存track
        self.device = settings['device']
        self.max_age = settings['max_age']
        self.min_hits = settings['min_hits']
        self.asso_func = ASSO_FUNCS[settings['asso_func']]
        self.delta_t = settings['delta_t']
        self.inertia = settings['inertia']
        self.iou_threshold = settings['iou_threshold']
        self.det_thresh = settings['det_thresh']#用以分辨高(>det_thresh)、中(0.1、det_thresh之間)、低分框(<0.1)
        self.pos_lambda = settings['pos_lambda']
        self.direction_type = settings['direction_type']
        self.count = 0
        self.dir_history_num = settings['dir_history_num']

        self.use_byte = settings['use_byte']                
        if self.use_byte:
            self.det_thresh_byte = settings['iou_threshold'] #低分框的threshold

        self.use_feature = settings['use_feature'] #是否要看特徵，否則純看IoU
        if self.use_feature: # strong_sort 的特徵
            self.samples = {} #用來存unique ID的特徵向量
            self.max_budget = settings['max_budget'] #每一個id可以儲存的特徵數量
            self.model = ReIDDetectMultiBackend(weights=settings['reid_weights'], device=self.device, fp16=settings['fp16'], pretrained=True) #ReID模型（特徵模型）
            img = torch.zeros((1, 3, 1920, 1080))
            features_len = len(self.model(img)[0])
            self.no_features = np.zeros((0, features_len))
            self.lamda = settings['lamda'] #cost appear_cost的權重
            self.appear_threshold = settings['appear_threshold'] # 特徵threshold
    
    def distance(self, features, targets): #計算"特徵"的距離(參考strong sort), 也可以使用歐式距離（但我沒寫), 以及我改寫原code的餘弦距離（改成愈大愈好，為了跟IoU分數加總）
        """Compute distance between features and targets.
        Parameters
        ----------
        features : ndarray
            An NxM matrix of N features of dimensionality M.
        targets : List[int]
            A list of targets to match the given `features` against.
        Returns
        -------
        ndarray
            Returns a cost matrix of shape len(targets), len(features), where
            element (i, j) contains the closest squared distance between
            `targets[i]` and `features[j]`.
        """
        # 結論：計算每一個當前bbox對應之前track的特徵距離計算（保留最大者的值)
        cost_matrix = np.zeros((len(targets), len(features)))
        for i, target in enumerate(targets):
            cost_matrix[i, :] = nn_cosine_distance(self.samples[target], features)
        #我需要 len(features), len(targets) #features=當前bbox, target=track_bbox, 故轉至
        #需要轉至 變成當前bbox對上track(因為我IoU匹配也是bbox對track)
        return cost_matrix.T 
        
    #更新每個ID的特徵 (strongsort)，並只考慮conf >= 0.7者，避免分數太低的特徵搗亂(遮擋)
    def partial_fit(self, features, targets, conf): 
        for i, (feature, target) in enumerate(zip(features, targets)):

            self.samples.setdefault(target, []).append(feature)
            if conf[i] < 0.7: #只保留高分框的特徵, 低分框會有雜質
                continue 
            if self.max_budget is not None: #避免存太多特徵量, 單一object最多只有max_budget個特徵儲存）
                self.samples[target] = self.samples[target][-self.max_budget:]
    
    def _get_features(self, bbox_xywh, ori_img): #抽取bbox特徵
        im_crops = []
        for box in bbox_xywh:
            x1, y1, x2, y2 = self._xywh_to_xyxy(box)
            im = ori_img[y1:y2, x1:x2]
            im_crops.append(im)

        if im_crops:
            features = self.model(im_crops)
        else:
            features = self.no_features.copy()
            
        return features
    
    def get_update_results(self):
        # if didn't update, return None
        if len(self.update_lst['new_ids']) == 0 and len(self.update_lst['disappeared_ids']) == 0:
            return None
        else:
            return self.update_lst
        
    def _xywh_to_xyxy(self, bbox_xywh):
        x, y, w, h = bbox_xywh
        x1 = max(int(x - w / 2), 0)
        x2 = min(int(x + w / 2), self.width - 1)
        y1 = max(int(y - h / 2), 0)
        y2 = min(int(y + h / 2), self.height - 1)
        
        return x1, y1, x2, y2
    
    def run(self, dets, ori_img): #KalmanFilter更新狀態位置
        """
        Params:
            dets - a numpy array of detections in the format [[x1,y1,x2,y2,score],[x1,y1,x2,y2,score],...]
        Requires: 
            this method must be called once for each frame even with empty detections (use np.empty((0, 5)) for frames without detections).
        Returns:
            the a similar array, where the last column is the object ID.
        """
        # init results
        # track_results = [-1 for i in range(len(dets))]
        track_results = np.ones(len(dets)) * (-1)
        speed_results = np.zeros(len(dets))
        direction_results = np.zeros((len(dets), 2)) if self.direction_type == 'vector' else np.zeros(len(dets))

        self.update_lst = {'new_ids': [], 'disappeared_ids': []}

        #先處理track的東西，避免沒有det到東西，track沒更新後就回傳，因此先處理track,再看有沒有det，來進行下一步
        # get predicted locations from existing trackers. trks用來儲存tracking

        # to_del = [] #若有刪除狀態的track, pop it    
        pred_results = [] #儲存pred_box
        trks = np.zeros((len(self.trackers), 6))
        #填滿trks
        for t, trk in enumerate(trks):
            #確認track狀態為confirmed才繼續
            pos = self.trackers[t].predict()[0] #拿上一幀的框先預測變成track, 以便後面匹配
            trk_id = self.trackers[t].id 
            trk[:] = [pos[0], pos[1], pos[2], pos[3], 0, trk_id+1] #多放id在後面

            pred_results.append([pos[0], pos[1], pos[2], pos[3]])
        trks = np.ma.compress_rows(np.ma.masked_invalid(trks))

        # 沒det，track需要更新空的，並且確認是否miss掉
        # if not(isinstance(dets, np.ndarray) and dets.shape[0]):
        if dets is None:
            if trks is not None:
                delete_list = []
                for m, trk in enumerate(trks):
                    self.trackers[m].update(None, None)
                    self.trackers[m].mark_missed() #檢查是否遺失太久， >= max_age -> deleted
                    if trk.is_deleted():
                        delete_list.append(m)
                        self.update_lst['disappeared_ids'].append(trk.id)
                for i in reversed(delete_list):
                    self.trackers.pop(i)

            if len(self.update_lst['disappeared_ids']) == 0:
                self.update_lst = None

            tracking_results = {
                'track_results': track_results,
                'speed_results': speed_results,
                'direction_results': direction_results,
                'feature_results': None
            }

            return tracking_results

        xyxys = dets[:, 0:4]
        confs = dets[:, 4]
        classes = dets[:, 5]
        #檢測bbox相對線之位置, 新增一個index-id（order to map the track
        index_lst = []
        for i in range(len(xyxys)):
            index_lst.append(i)

        # 抽取目前bbox特徵(新增的--從strongsort copy過來，並stack起來)
        self.height, self.width = ori_img.shape[:2]
        xywhs = xyxy2xywh(xyxys)
        if self.use_feature: #若有考慮特徵, 則先擷取特徵, 並stack起來
            feature = self._get_features(xywhs, ori_img)
            if isinstance(feature, torch.Tensor):
                features = feature.cpu().detach().numpy()
            elif isinstance(feature, np.ndarray):
                features = feature

            # [0:4] = bbox, [4]:confs [5]:classes [6:]det_results_index [7:] features
            output_results = np.column_stack((xyxys, confs, classes, index_lst, features))
        else:
            output_results = np.column_stack((xyxys, confs, classes, index_lst))
        
        # 將detection分成三等分（高、中、低）
        inds_low = confs > 0.1
        inds_middle = confs < self.det_thresh
        inds_second = np.logical_and(inds_low, inds_middle)  # self.det_thresh > score > 0.1, for second matching
        
        #高分框
        inds_high = confs >= self.det_thresh
        dets_high = output_results[inds_high]  #高分框(若有use_byte則是det_thresh, 否則是conf_thres)
        #中分框
        if self.use_byte: #若有use_byte:考慮中分框，先存起來以最後initial方便
            dets_second = output_results[inds_second]  # detections for second matching  
            unmatched_dets_second = []
            for i in range(len(dets_second)):
                unmatched_dets_second.append(i)
        
        b_features = [] # b_features暫存feature_list
        velocities = np.array(
            [trk.velocity if trk.velocity is not None else np.array((0, 0)) for trk in self.trackers])
        last_boxes = np.array([trk.last_observation for trk in self.trackers])
        k_observations = np.array(
            [k_previous_obs(trk.observations, trk.age, self.delta_t) for trk in self.trackers])
        """
            First round of association 先針對高分框, 進行匹配,（若有use_feature, 則combine一起配對, 否則單純看IoU)
        """ 
        matching_table =[]
        matched, unmatched_dets, unmatched_trks = associate(self,
            dets_high, trks, velocities, k_observations, self.inertia, self.use_feature)
        if len(matched) != 0:
            for det_index, track_index in matched:
                matching_table.append([int(dets_high[det_index][6]), track_index])
        for m in matched:
            self.trackers[m[1]].update(dets_high[m[0], :], dets_high[m[0], 5])

        """
            Second round of associaton by OCR (下面都只考慮bbox的IoU匹配分數)
            可以調整的有assoc_func(也就是bbox的IoU方式選擇:giou、ciou、diou、iou、ct_dist)
            (1)若use_feature :  
                1.第二階段先將剩餘的高分框unmatched_det與unmatched_trks匹配！(避免有高分框因特徵太相似而被忽略)
                2.再將剩餘的unmatched_trks與中分框匹配
            (2)else:
                1.由於第一階段單純只考慮IoU，故接下來先將中分框與unmatched_trks匹配
                2.再將剩下的unmatched_trks與第一階段沒匹配的unmatched_det匹配                    
        """
        # 針對剩下的unmatched_trks和高分框的unmatched_dets「進行單純IoU匹配」
        if self.use_feature:
            if unmatched_dets.shape[0] > 0 and unmatched_trks.shape[0] > 0:
                left_dets = dets_high[unmatched_dets]
                # left_trks = last_boxes[unmatched_trks]
                left_trks = trks[unmatched_trks]
                
                iou_left = self.asso_func(left_dets, left_trks)
                iou_left = np.array(iou_left)
                if iou_left.max() > self.iou_threshold:  
                    """
                        NOTE: by using a lower threshold, e.g., self.iou_threshold - 0.1, you may
                        get a higher performance especially on MOT17/MOT20 datasets. But we keep it
                        uniform here for simplicity
                    """
                    rematched_indices = linear_assignment(-iou_left)
                    to_remove_det_indices = []
                    to_remove_trk_indices = []
                    for m in rematched_indices:
                        det_ind, trk_ind = unmatched_dets[m[0]], unmatched_trks[m[1]]
                        matching_table.append([int(dets_high[det_ind][6]), trk_ind])

                        if iou_left[m[0], m[1]] < self.iou_threshold:
                            continue
                        self.trackers[trk_ind].update(dets_high[det_ind, :], dets_high[det_ind, 5])

                        to_remove_det_indices.append(det_ind)
                        to_remove_trk_indices.append(trk_ind)

                    unmatched_dets = np.setdiff1d(unmatched_dets, np.array(to_remove_det_indices))
                    unmatched_trks = np.setdiff1d(unmatched_trks, np.array(to_remove_trk_indices))

                    # if len(rematched_indices) != 0:
                    #     for det_index, track_index in rematched_indices:
                    #         matching_table.append([int(dets_high[det_index][7]), unmatched_trks[track_index]])

            # BYTE association 針對中分框, 且有un_matched_track時，若iou重合 > threshold ---> 進行匹配
            if self.use_byte and len(dets_second) > 0 and unmatched_trks.shape[0] > 0:
                u_trks = trks[unmatched_trks]
                iou_left = self.asso_func(dets_second, u_trks)          # iou between low score detections and unmatched tracks
                iou_left = np.array(iou_left)
                if iou_left.max() > self.det_thresh_byte:
                    """
                        NOTE: by using a lower threshold, e.g., self.iou_threshold - 0.1, you may
                        get a higher performance especially on MOT17/MOT20 datasets. But we keep it
                        uniform here for simplicity
                    """
                    matched_indices = linear_assignment(-iou_left)
                    to_remove_trk_indices = []
                    to_remove_detsecond_indices = []
                    for m in matched_indices:
                        det_ind, trk_ind = m[0], unmatched_trks[m[1]]
                        matching_table.append([int(dets_second[det_ind][6]), trk_ind])

                        if iou_left[m[0], m[1]] < self.det_thresh_byte:
                            continue
                        self.trackers[trk_ind].update(dets_second[det_ind, :], dets_second[det_ind, 5])
     
                        to_remove_trk_indices.append(trk_ind)
                        to_remove_detsecond_indices.append(det_ind)

                    unmatched_dets_second = np.setdiff1d(unmatched_dets_second, np.array(to_remove_detsecond_indices))
                    unmatched_trks = np.setdiff1d(unmatched_trks, np.array(to_remove_trk_indices))
            
                # if len(matched_indices) != 0:
                #     for det_index, track_index in matched_indices:
                #         matching_table.append([int(dets_second[det_index][7]), unmatched_trks[track_index]])
        else:
            # BYTE association 針對中分框, 且有un_matched_track時，若iou重合 > threshold ---> 進行匹配
            if self.use_byte and len(dets_second) > 0 and unmatched_trks.shape[0] > 0:
                u_trks = trks[unmatched_trks]
                iou_left = self.asso_func(dets_second, u_trks)          # iou between low score detections and unmatched tracks
                iou_left = np.array(iou_left)
                if iou_left.max() > self.det_thresh_byte:
                    """
                        NOTE: by using a lower threshold, e.g., self.iou_threshold - 0.1, you may
                        get a higher performance especially on MOT17/MOT20 datasets. But we keep it
                        uniform here for simplicity
                    """
                    matched_indices = linear_assignment(-iou_left)
                    # matched_indices = linear_assignment(-cost_matrix)
                    to_remove_trk_indices = []
                    to_remove_detsecond_indices = []
                    for m in matched_indices:
                        det_ind, trk_ind = m[0], unmatched_trks[m[1]]
                        matching_table.append([int(dets_second[det_ind][6]), trk_ind])

                        if iou_left[m[0], m[1]] < self.det_thresh_byte:
                            continue
                        
                        self.trackers[trk_ind].update(dets_second[det_ind, :], dets_second[det_ind, 5])

                        to_remove_trk_indices.append(trk_ind)
                        to_remove_detsecond_indices.append(det_ind)
                        
                    unmatched_dets_second = np.setdiff1d(unmatched_dets_second, np.array(to_remove_detsecond_indices))
                    unmatched_trks = np.setdiff1d(unmatched_trks, np.array(to_remove_trk_indices))

            if unmatched_dets.shape[0] > 0 and unmatched_trks.shape[0] > 0:
                left_dets = dets_high[unmatched_dets]
                left_trks = last_boxes[unmatched_trks]
                # left_trks = trks[unmatched_trks]
                iou_left = self.asso_func(left_dets, left_trks)
                iou_left = np.array(iou_left)
                if iou_left.max() > self.iou_threshold:  
                    """
                        NOTE: by using a lower threshold, e.g., self.iou_threshold - 0.1, you may
                        get a higher performance especially on MOT17/MOT20 datasets. But we keep it
                        uniform here for simplicity
                    """
                    rematched_indices = linear_assignment(-iou_left)
                    to_remove_det_indices = []
                    to_remove_trk_indices = []
                    for m in rematched_indices:
                        det_ind, trk_ind = unmatched_dets[m[0]], unmatched_trks[m[1]]
                        matching_table.append([int(dets_high[det_ind][6]), trk_ind])

                        if iou_left[m[0], m[1]] < self.iou_threshold:
                            continue

                        self.trackers[trk_ind].update(dets_high[det_ind, :], dets_high[det_ind, 5])

                        
                        to_remove_det_indices.append(det_ind)
                        to_remove_trk_indices.append(trk_ind)
                    
                    unmatched_dets = np.setdiff1d(unmatched_dets, np.array(to_remove_det_indices))
                    unmatched_trks = np.setdiff1d(unmatched_trks, np.array(to_remove_trk_indices))

        for m in unmatched_trks:
            self.trackers[m].update(None, None)
            self.trackers[m].mark_missed() #檢查是否遺失太久
            
        # create and initialise new trackers for unmatched detections
        for i in unmatched_dets:
            trk = KalmanBoxTracker(dets_high[i, :], self.max_age , self.min_hits, dets_high[i, 5], delta_t=self.delta_t, use_feature = self.use_feature, image_size = (self.width, self.height), id = self.count, dir_history_num=self.dir_history_num)
            self.count += 1
            self.trackers.append(trk)
            self.update_lst['new_ids'].append(trk.id)

        if self.use_byte: #針對use_byte, 還需對中分框裡匹配到的新det進行initial
            for i in unmatched_dets_second:
                trk = KalmanBoxTracker(dets_second[i, :], self.max_age , self.min_hits, dets_second[i, 5], delta_t=self.delta_t, use_feature = self.use_feature, image_size = (self.width, self.height), id = self.count, dir_history_num=self.dir_history_num)
                self.count += 1
                self.trackers.append(trk)
                self.update_lst['new_ids'].append(trk.id)
                
        if len(matching_table) != 0:
            for det_index,track_index in matching_table:
                trk = self.trackers[track_index]
                if trk.is_confirmed():
                    track_results[det_index] = trk.id

                    last_time = trk.time
                    cur_time = time.time()
                    speed, direction = get_speed_direction(last_time,
                                                           cur_time,
                                                           pre_box=trk.history_observations[-2][:4],
                                                           pre_n_box=trk.history_observations[0][:4],
                                                           cur_box=dets[det_index, :4],
                                                           pos_lambda=self.pos_lambda,
                                                           direction_type=self.direction_type)

                    trk.time = cur_time
                    trk.speed = speed
                    trk.direction = direction
                    speed_results[det_index] = trk.speed
                    direction_results[det_index] = trk.direction

        delete_list=[]
        for i, trk in enumerate(self.trackers):
            if trk.last_observation.sum() < 0:                
                d = trk.get_state()[0]
            else:
                """
                    this is optional to use the recent observation or the kalman filter prediction,
                    we didn't notice significant difference here
                """
                d = trk.last_observation[:4] #前一幀的結果x

            if self.use_feature:
                b_features.append(np.concatenate((d, [trk.id+1], [trk.cls], [trk.conf], trk.feature)).reshape(1, -1))
            # remove dead tracklet
            if trk.is_deleted():
                delete_list.append(i)
                self.update_lst['disappeared_ids'].append(trk.id)

        for i in reversed(delete_list):
            self.trackers.pop(i)
            
        # partial_fit用以記錄當前box_id的特徵
        if len(b_features) > 0:
            if self.use_feature :
                box = np.concatenate(b_features)
                b_features = box[:, 7:] 
                b_id = box[:,4]
                b_conf = box[:,6]
                #最後把這次配對到的ID的特徵更新
                if self.device == 'cuda':
                    self.partial_fit(np.asarray(b_features), np.asarray(b_id), np.asarray(b_conf))
                else:
                    self.partial_fit(np.asarray(b_features), np.asarray(b_id), np.asarray(b_conf))

        tracking_results = {
            'track_ids': track_results,
            'speed_results': speed_results,
            'direction_results': direction_results,
            'feature_results': features if self.use_feature else None
        }

        return tracking_results



import numpy as np
import time
from PIL import Image, ImageDraw, ImageFont


def _cosine_distance(a, b, data_is_normalized=False):
    """Compute pair-wise cosine distance between points in `a` and `b`.
    Parameters
    ----------
    a : array_like
        An NxM matrix of N samples of dimensionality M.
    b : array_like
        An LxM matrix of L samples of dimensionality M.
    data_is_normalized : Optional[bool]
        If True, assumes rows in a and b are unit length vectors.
        Otherwise, a and b are explicitly normalized to lenght 1.
    Returns
    -------
    ndarray
        Returns a matrix of size len(a), len(b) such that eleement (i, j)
        contains the squared distance between `a[i]` and `b[j]`.
    """
    if not data_is_normalized:
        a = np.asarray(a) / np.linalg.norm(a, axis=1, keepdims=True)
        b = np.asarray(b) / np.linalg.norm(b, axis=1, keepdims=True)
        
    return 1. - np.dot(a, b.T)

def cosine_distance_to_image_center(bbox, image_size):
    """
    Calculates the absolute cosine distance subtracted by 1, from an up vector located at the image center to
    the specified bbox.

    Inputs:
        bbox - bounding box of type (x pos, y pos, s area, r aspect ratio), shape (1, 4)
        image_size - tuple (width, height)

    Returns:
        Absolute value of cosine distance subtracted by 1 (higher values means further away from up vector located
        at image center)
    """
    up_vector = np.asarray([[0., 1., 0., 0.]])
    image_center_pos = np.asarray([[image_size[0] / 2., image_size[1] / 2., 0., 0.]])
    
    return np.abs(_cosine_distance(up_vector, bbox - image_center_pos).flatten() - 1)[0]

def speed_direction(bbox1, bbox2):
    cx1, cy1 = (bbox1[0]+bbox1[2]) / 2.0, (bbox1[1]+bbox1[3])/2.0
    cx2, cy2 = (bbox2[0]+bbox2[2]) / 2.0, (bbox2[1]+bbox2[3])/2.0
    speed = np.array([cy2-cy1, cx2-cx1])
    norm = np.sqrt((cy2-cy1)**2 + (cx2-cx1)**2) + 1e-6
    
    return speed / norm

def convert_bbox_to_z(bbox):
    """
    Takes a bounding box in the form [x1,y1,x2,y2] and returns z in the form
      [x,y,s,r] where x,y is the centre of the box and s is the scale/area and r is
      the aspect ratio
    """
    w = bbox[2] - bbox[0]
    h = bbox[3] - bbox[1]
    x = bbox[0] + w/2.
    y = bbox[1] + h/2.
    s = w * h  # scale is just area
    r = w / float(h+1e-6)
    
    return np.array([x, y, s, r]).reshape((4, 1))

def convert_x_to_bbox(x, score=None):
    """
    Takes a bounding box in the centre form [x,y,s,r] and returns it in the form
      [x1,y1,x2,y2] where x1,y1 is the top left and x2,y2 is the bottom right
    """
    w = np.sqrt(x[2] * x[3])
    h = x[2] / w
    if(score == None):
      return np.array([x[0]-w/2., x[1]-h/2., x[0]+w/2., x[1]+h/2.]).reshape((1, 4))
    else:
      return np.array([x[0]-w/2., x[1]-h/2., x[0]+w/2., x[1]+h/2., score]).reshape((1, 5))

#紀錄track的狀態----超過min_hits才變confirm！（避免錯誤偵測也追蹤）
class TrackState:

    initial = 1
    Confirmed = 2
    Deleted = 3

class KalmanBoxTracker(object):
    """
    This class represents the internal state of individual tracked objects observed as bbox.
    """
    count = 0
    def __init__(self, bbox, max_age, min_hits, cls, delta_t=3, orig=False, use_feature = False, image_size=(0, 0), id = 0, dir_history_num=10): 
        self.image_size = image_size
        # box ---> [0:4] = bbox, [4]:confs [5]:classes [6:] positions [7:] features
        """
        Initialises a tracker using initial bounding box.

        """
        # define constant velocity model
        if not orig:
          from .kalmanfilter import KalmanFilterNew as KalmanFilter
          self.kf = KalmanFilter(dim_x=7, dim_z=4)
        # else:
        #   from filterpy.kalman import KalmanFilter
        #   self.kf = KalmanFilter(dim_x=7, dim_z=4)        
        '''
        testing
        '''
        
        self.kf.F = np.array([
            [1, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 0, 1, 0],
            [0, 0, 1, 0, 0, 0, 1],
            [0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 1]])
        self.kf.H = np.array([
            [1, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0]])

        """
            Motion and observation uncertainty are chosen relative to the current
            state estimate. These weights control the amount of uncertainty in
            the model. This is a bit hacky.
        """
        self._std_weight_position = 10.
        self._std_weight_size = 1. / 20
        self._std_weight_velocity = 20.

        state_space_bbox = convert_bbox_to_z(bbox)

        dist = cosine_distance_to_image_center(state_space_bbox, self.image_size)
        std = [
            2 * self._std_weight_position * dist + 1,  # the center point x
            2 * self._std_weight_position * dist + 1,  # the center point y
            2 * self._std_weight_size * state_space_bbox[2][0],  # the scale (area)
            1 * state_space_bbox[3][0],  # the ratio of width/height
            10 * self._std_weight_velocity * dist + 1,
            10 * self._std_weight_velocity * dist + 1,
            10 * self._std_weight_velocity * dist + 1]
        self.kf.P = np.diag(np.square(std).flatten())

        self.kf.x[:4] = state_space_bbox
        '''
        testing
        '''
        
        # self.kf.R[2:, 2:] *= 10.
        # self.kf.P[4:, 4:] *= 1000.  # give high uncertainty to the unobservable initial velocities
        # self.kf.P *= 10.
        # self.kf.Q[-1, -1] *= 0.01
        # self.kf.Q[4:, 4:] *= 0.01

        # self.kf.x[:4] = convert_bbox_to_z(bbox[:4])
        self.time_since_update = 0
        # self.id = KalmanBoxTracker.count
        # KalmanBoxTracker.count += 1
        self.id = id
        
        self.history = []
        self.hits = 0
        self.hit_streak = 0
        self.age = 0
        self.conf = bbox[4]
        if use_feature:
            self.feature = bbox[7:]#特徵
        self.tmp_index = bbox[6] #for mapping det id
        self.last_pos = None #上一期的pos
        
        self.cls = cls
        self._n_hits = min_hits #最少hit幾次就confirm
        self._max_age = max_age #confirm track多久後沒出現就delete
        self.state = TrackState.initial #紀錄track當前狀態
        self.use_feature = use_feature #使用外觀特徵or_not
        self.count_flag = False #記錄是否被累加過，針對人流、車流、進來的總人數or出去的總人數做控管
        """
        NOTE: [-1,-1,-1,-1,-1] is a compromising placeholder for non-observation status, the same for the return of 
        function k_previous_obs. It is ugly and I do not like it. But to support generate observation array in a 
        fast and unified way, which you would see below k_observations = np.array([k_previous_obs(...]]), let's bear it for now.
        """
        self.last_observation = np.array([-1, -1, -1, -1, -1])  # placeholder
        self.observations = dict()
        self.history_observations = []
        self.velocity = None
        self.delta_t = delta_t
        self.dir_history_num = dir_history_num

        self.time = time.time()
        self.speed = 0
        self.direction = None


    def update(self, bbox, cls):
        # [0:4] = bbox, [4]:confs [5]:classes [6:] input_index [7:] features
        """
        Updates the state vector with observed bbox.
        """
        
        if bbox is not None:
            self.conf = bbox[4]
            self.cls = cls
            if self.use_feature:
                self.feature = bbox[7:]
            if self.last_observation.sum() >= 0:  # no previous observation
                previous_box = None
                for i in range(self.delta_t):
                    dt = self.delta_t - i
                    if self.age - dt in self.observations:
                        previous_box = self.observations[self.age-dt]
                        break
                if previous_box is None:
                    previous_box = self.last_observation
                """
                  Estimate the track speed direction with observations \Delta t steps away
                """
                self.velocity = speed_direction(previous_box, bbox)
            
            
            """
            Update covariance matrices
            """
            dist = cosine_distance_to_image_center(bbox[:4].reshape(1, -1), self.image_size)
            std = [
                2. * self._std_weight_position * dist + 1,
                2. * self._std_weight_position * dist + 1,
                self._std_weight_size * self.kf.x[2][0],
                1 * self.kf.x[3][0]]
            std = [(2 - self.conf) * x for x in std]
            self.kf.R = np.diag(np.square(std).flatten())
            
            
            
            """
              Insert new observations. This is a ugly way to maintain both self.observations
              and self.history_observations. Bear it for the moment.
            """
            self.last_observation = bbox[:5]
            self.observations[self.age] = bbox[:5]
            self.history_observations.append(bbox[:5])
            if len(self.history_observations) > self.dir_history_num:
                self.history_observations = self.history_observations[-self.dir_history_num-1:] # is a ugly way for same with simple track!!

            self.time_since_update = 0
            self.hits += 1
            if self.state == TrackState.initial and self.hits >= self._n_hits:
                self.state = TrackState.Confirmed
            self.history = []
            self.hit_streak += 1
            self.kf.update(convert_bbox_to_z(bbox[:5]))

        else:
            self.kf.update(bbox)

    def predict(self):
        """
        Advances the state vector and returns the predicted bounding box estimate.
        """
        if((self.kf.x[6]+self.kf.x[2]) <= 0):
            self.kf.x[6] *= 0.0
        '''
        test
        '''
        # Update process noise
        dist = cosine_distance_to_image_center(self.kf.x, self.image_size)
        std_pos = [
            self._std_weight_position * dist + 1,
            self._std_weight_position * dist + 1,
            self._std_weight_size * self.kf.x[2][0],
            1 * self.kf.x[3][0]]
        std_vel = [
            self._std_weight_velocity * dist + 1,
            self._std_weight_velocity * dist + 1,
            self._std_weight_velocity * dist + 1]
        self.kf.Q = np.diag(np.square(np.r_[std_pos, std_vel]).flatten())
        '''
        test
        '''
        self.kf.predict()
        self.age += 1
        if(self.time_since_update > 0):
            self.hit_streak = 0
        self.time_since_update += 1
        self.history.append(convert_x_to_bbox(self.kf.x))
        return self.history[-1]


    def get_state(self):
        """
        Returns the current bounding box estimate.
        """
        return convert_x_to_bbox(self.kf.x)
    
    def mark_missed(self):
        """Mark this track as missed (no association at the current time step).
        """
        if self.state == TrackState.initial:
            self.state = TrackState.Deleted
        elif self.time_since_update > self._max_age:
            self.state = TrackState.Deleted

    def is_initial(self):
        """Returns True if this track is tentative (unconfirmed).
        """
        return self.state == TrackState.initial

    def is_confirmed(self):
        """Returns True if this track is confirmed."""
        return self.state == TrackState.Confirmed

    def is_deleted(self):
        """Returns True if this track is dead and should be deleted."""
        return self.state == TrackState.Deleted    